#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Models representing flavors."""

from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional

from pydantic import Field

from zenml.constants import STR_FIELD_MAX_LENGTH
from zenml.enums import StackComponentType
from zenml.models.v2.base.base import BaseUpdate
from zenml.models.v2.base.scoped import (
    UserScopedFilter,
    UserScopedRequest,
    UserScopedResponse,
    UserScopedResponseBody,
    UserScopedResponseMetadata,
    UserScopedResponseResources,
)

if TYPE_CHECKING:
    from zenml.models import (
        ServiceConnectorRequirements,
    )

# ------------------ Request Model ------------------


class FlavorRequest(UserScopedRequest):
    """Request model for stack component flavors."""

    ANALYTICS_FIELDS: ClassVar[List[str]] = [
        "type",
        "integration",
    ]

    name: str = Field(
        title="The name of the Flavor.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    display_name: Optional[str] = Field(
        default=None,
        title="The display name of the Flavor.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    type: StackComponentType = Field(title="The type of the Flavor.")
    config_schema: Dict[str, Any] = Field(
        title="The JSON schema of this flavor's corresponding configuration.",
    )
    connector_type: Optional[str] = Field(
        default=None,
        title="The type of the connector that this flavor uses.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    connector_resource_type: Optional[str] = Field(
        default=None,
        title="The resource type of the connector that this flavor uses.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    connector_resource_id_attr: Optional[str] = Field(
        default=None,
        title="The name of an attribute in the stack component configuration "
        "that plays the role of resource ID when linked to a service "
        "connector.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    source: str = Field(
        title="The path to the module which contains this Flavor.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    integration: Optional[str] = Field(
        title="The name of the integration that the Flavor belongs to.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    logo_url: Optional[str] = Field(
        default=None,
        title="Optionally, a url pointing to a png,"
        "svg or jpg can be attached.",
    )
    docs_url: Optional[str] = Field(
        default=None,
        title="Optionally, a url pointing to docs, within docs.zenml.io.",
    )
    sdk_docs_url: Optional[str] = Field(
        default=None,
        title="Optionally, a url pointing to SDK docs,"
        "within sdkdocs.zenml.io.",
    )
    is_custom: bool = Field(
        title="Whether or not this flavor is a custom, user created flavor.",
        default=True,
    )


# ------------------ Update Model ------------------


class FlavorUpdate(BaseUpdate):
    """Update model for stack component flavors."""

    name: Optional[str] = Field(
        title="The name of the Flavor.",
        max_length=STR_FIELD_MAX_LENGTH,
        default=None,
    )
    display_name: Optional[str] = Field(
        title="The display name of the Flavor.",
        max_length=STR_FIELD_MAX_LENGTH,
        default=None,
    )
    type: Optional[StackComponentType] = Field(
        title="The type of the Flavor.", default=None
    )
    config_schema: Optional[Dict[str, Any]] = Field(
        title="The JSON schema of this flavor's corresponding configuration.",
        default=None,
    )
    connector_type: Optional[str] = Field(
        title="The type of the connector that this flavor uses.",
        max_length=STR_FIELD_MAX_LENGTH,
        default=None,
    )
    connector_resource_type: Optional[str] = Field(
        title="The resource type of the connector that this flavor uses.",
        max_length=STR_FIELD_MAX_LENGTH,
        default=None,
    )
    connector_resource_id_attr: Optional[str] = Field(
        title="The name of an attribute in the stack component configuration "
        "that plays the role of resource ID when linked to a service "
        "connector.",
        max_length=STR_FIELD_MAX_LENGTH,
        default=None,
    )
    source: Optional[str] = Field(
        title="The path to the module which contains this Flavor.",
        max_length=STR_FIELD_MAX_LENGTH,
        default=None,
    )
    integration: Optional[str] = Field(
        title="The name of the integration that the Flavor belongs to.",
        max_length=STR_FIELD_MAX_LENGTH,
        default=None,
    )
    logo_url: Optional[str] = Field(
        title="Optionally, a url pointing to a png,"
        "svg or jpg can be attached.",
        default=None,
    )
    docs_url: Optional[str] = Field(
        title="Optionally, a url pointing to docs, within docs.zenml.io.",
        default=None,
    )
    sdk_docs_url: Optional[str] = Field(
        title="Optionally, a url pointing to SDK docs,"
        "within sdkdocs.zenml.io.",
        default=None,
    )
    is_custom: Optional[bool] = Field(
        title="Whether or not this flavor is a custom, user created flavor.",
        default=None,
    )


# ------------------ Response Model ------------------


class FlavorResponseBody(UserScopedResponseBody):
    """Response body for stack component flavors."""

    type: StackComponentType = Field(title="The type of the Flavor.")
    display_name: str = Field(
        title="The display name of the Flavor.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    integration: Optional[str] = Field(
        title="The name of the integration that the Flavor belongs to.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    source: str = Field(
        title="The path to the module which contains this Flavor.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    logo_url: Optional[str] = Field(
        default=None,
        title="Optionally, a url pointing to a png,"
        "svg or jpg can be attached.",
    )
    is_custom: bool = Field(
        title="Whether or not this flavor is a custom, user created flavor.",
        default=True,
    )


class FlavorResponseMetadata(UserScopedResponseMetadata):
    """Response metadata for stack component flavors."""

    config_schema: Dict[str, Any] = Field(
        title="The JSON schema of this flavor's corresponding configuration.",
    )
    connector_type: Optional[str] = Field(
        default=None,
        title="The type of the connector that this flavor uses.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    connector_resource_type: Optional[str] = Field(
        default=None,
        title="The resource type of the connector that this flavor uses.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    connector_resource_id_attr: Optional[str] = Field(
        default=None,
        title="The name of an attribute in the stack component configuration "
        "that plays the role of resource ID when linked to a service "
        "connector.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    docs_url: Optional[str] = Field(
        default=None,
        title="Optionally, a url pointing to docs, within docs.zenml.io.",
    )
    sdk_docs_url: Optional[str] = Field(
        default=None,
        title="Optionally, a url pointing to SDK docs,"
        "within sdkdocs.zenml.io.",
    )


class FlavorResponseResources(UserScopedResponseResources):
    """Response resources for stack component flavors."""


class FlavorResponse(
    UserScopedResponse[
        FlavorResponseBody,
        FlavorResponseMetadata,
        FlavorResponseResources,
    ]
):
    """Response model for stack component flavors."""

    # Analytics
    ANALYTICS_FIELDS: ClassVar[List[str]] = [
        "id",
        "type",
        "integration",
    ]

    name: str = Field(
        title="The name of the Flavor.",
        max_length=STR_FIELD_MAX_LENGTH,
    )

    def get_hydrated_version(self) -> "FlavorResponse":
        """Get the hydrated version of the flavor.

        Returns:
            an instance of the same entity with the metadata field attached.
        """
        from zenml.client import Client

        return Client().zen_store.get_flavor(self.id)

    # Helper methods
    @property
    def connector_requirements(
        self,
    ) -> Optional["ServiceConnectorRequirements"]:
        """Returns the connector requirements for the flavor.

        Returns:
            The connector requirements for the flavor.
        """
        from zenml.models import (
            ServiceConnectorRequirements,
        )

        if not self.connector_resource_type:
            return None

        return ServiceConnectorRequirements(
            connector_type=self.connector_type,
            resource_type=self.connector_resource_type,
            resource_id_attr=self.connector_resource_id_attr,
        )

    # Body and metadata properties
    @property
    def type(self) -> StackComponentType:
        """The `type` property.

        Returns:
            the value of the property.
        """
        return self.get_body().type

    @property
    def display_name(self) -> str:
        """The `display_name` property.

        Returns:
            the value of the property.
        """
        return self.get_body().display_name

    @property
    def integration(self) -> Optional[str]:
        """The `integration` property.

        Returns:
            the value of the property.
        """
        return self.get_body().integration

    @property
    def source(self) -> str:
        """The `source` property.

        Returns:
            the value of the property.
        """
        return self.get_body().source

    @property
    def logo_url(self) -> Optional[str]:
        """The `logo_url` property.

        Returns:
            the value of the property.
        """
        return self.get_body().logo_url

    @property
    def is_custom(self) -> bool:
        """The `is_custom` property.

        Returns:
            the value of the property.
        """
        return self.get_body().is_custom

    @property
    def config_schema(self) -> Dict[str, Any]:
        """The `config_schema` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().config_schema

    @property
    def connector_type(self) -> Optional[str]:
        """The `connector_type` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().connector_type

    @property
    def connector_resource_type(self) -> Optional[str]:
        """The `connector_resource_type` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().connector_resource_type

    @property
    def connector_resource_id_attr(self) -> Optional[str]:
        """The `connector_resource_id_attr` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().connector_resource_id_attr

    @property
    def docs_url(self) -> Optional[str]:
        """The `docs_url` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().docs_url

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """The `sdk_docs_url` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().sdk_docs_url


# ------------------ Filter Model ------------------


class FlavorFilter(UserScopedFilter):
    """Model to enable advanced stack component flavor filtering."""

    name: Optional[str] = Field(
        default=None,
        description="Name of the flavor",
    )
    display_name: Optional[str] = Field(
        default=None,
        description="Display name of the flavor",
    )
    type: Optional[str] = Field(
        default=None,
        description="Stack Component Type of the stack flavor",
    )
    integration: Optional[str] = Field(
        default=None,
        description="Integration associated with the flavor",
    )



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""FastAPI adapter implementations."""

from typing import Any, Callable, Dict, Optional

from fastapi import APIRouter, Depends, FastAPI, HTTPException
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from fastapi.staticfiles import StaticFiles
from starlette.middleware.base import BaseHTTPMiddleware

from zenml.config import (
    EndpointMethod,
    EndpointSpec,
    MiddlewareSpec,
)
from zenml.deployers.server.adapters import (
    EndpointAdapter,
    MiddlewareAdapter,
)
from zenml.deployers.server.app import BaseDeploymentAppRunner


class FastAPIEndpointAdapter(EndpointAdapter):
    """FastAPI implementation of endpoint adapter."""

    def _build_auth_dependency(self, api_key: str) -> Callable[..., Any]:
        """Build a FastAPI auth dependency.

        Args:
            api_key: The API key to use for authentication.

        Returns:
            FastAPI auth enforcement callable.
        """
        security = HTTPBearer(
            scheme_name="Bearer Token",
            description="Enter your API key as a Bearer token",
            auto_error=False,
        )

        def verify_token(
            credentials: Optional[HTTPAuthorizationCredentials] = Depends(
                security
            ),
        ) -> None:
            """Verify the provided Bearer token for authentication.

            Args:
                credentials: HTTP Bearer credentials from the request.

            Raises:
                HTTPException: If token is invalid.
            """
            if not credentials:
                raise HTTPException(
                    status_code=401,
                    detail="Authorization header required",
                    headers={"WWW-Authenticate": "Bearer"},
                )
            if credentials.credentials != api_key:
                raise HTTPException(
                    status_code=401,
                    detail="Invalid authentication token",
                    headers={"WWW-Authenticate": "Bearer"},
                )

        return verify_token

    def register_endpoint(
        self,
        app_runner: BaseDeploymentAppRunner,
        spec: EndpointSpec,
    ) -> None:
        """Register endpoint with FastAPI.

        Args:
            app_runner: Deployment app runner instance.
            spec: Framework-agnostic endpoint specification.

        Raises:
            RuntimeError: If the adapter is not used with a FastAPI application.
        """
        app = app_runner.asgi_app

        if not isinstance(app, FastAPI):
            raise RuntimeError(
                f"The {self.__class__.__name__} adapter must be used with a "
                "FastAPI application"
            )

        # Ensure handler is loaded
        handler = self.resolve_endpoint_handler(app_runner, spec)

        # Apply auth dependency if required
        dependencies = []
        if spec.auth_required and app_runner.deployment.auth_key:
            auth_dependency = self._build_auth_dependency(
                app_runner.deployment.auth_key
            )
            dependencies.append(Depends(auth_dependency))

        if spec.native:
            if isinstance(handler, APIRouter):
                app.include_router(
                    handler, prefix=spec.path, **spec.extra_kwargs
                )
                return
            # Handle StaticFiles mounting
            if isinstance(handler, StaticFiles):
                app.mount(spec.path, handler, name=spec.path.strip("/"))
                return

        # Register with appropriate HTTP method
        route_kwargs: Dict[str, Any] = {"dependencies": dependencies}
        route_kwargs.update(spec.extra_kwargs)

        if spec.method == EndpointMethod.GET:
            app.get(spec.path, **route_kwargs)(handler)
        elif spec.method == EndpointMethod.POST:
            app.post(spec.path, **route_kwargs)(handler)
        elif spec.method == EndpointMethod.PUT:
            app.put(spec.path, **route_kwargs)(handler)
        elif spec.method == EndpointMethod.PATCH:
            app.patch(spec.path, **route_kwargs)(handler)
        elif spec.method == EndpointMethod.DELETE:
            app.delete(spec.path, **route_kwargs)(handler)


class FastAPIMiddlewareAdapter(MiddlewareAdapter):
    """FastAPI implementation of middleware adapter.

    We support two types of native middleware:

    * A middleware class like that receives the ASGIApp object in the
    constructor and implements the __call__ method to dispatch the middleware,
    e.g.:

    ```python
    from starlette.types import ASGIApp, Receive, Scope, Send

    class MyMiddleware:
        def __init__(self, app: ASGIApp) -> None:
            self.app = app

        async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
            ...
            await self.app(scope, receive, send)
    ```

    * A middleware function that takes request and next callable and returns a response,
    e.g.:

    ```python
    from fastapi import Request, Response

    async def my_middleware(request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
        ...
        return await call_next(request)
    ```

    """

    def register_middleware(
        self,
        app_runner: BaseDeploymentAppRunner,
        spec: MiddlewareSpec,
    ) -> None:
        """Register middleware with FastAPI.

        Args:
            app_runner: Deployment app runner instance.
            spec: Framework-agnostic middleware specification.

        Raises:
            RuntimeError: If the adapter is not used with a FastAPI application.
        """
        app = app_runner.asgi_app

        if not isinstance(app, FastAPI):
            raise RuntimeError(
                f"The {self.__class__.__name__} adapter must be used with a "
                "FastAPI application"
            )

        middleware = self.resolve_middleware_handler(app_runner, spec)

        if spec.native:
            if isinstance(middleware, type):
                app.add_middleware(
                    middleware,  # type: ignore[arg-type]
                    **spec.extra_kwargs,
                )
                return

            app.add_middleware(
                BaseHTTPMiddleware,
                dispatch=middleware,
                **spec.extra_kwargs,
            )

        if isinstance(middleware, type):
            app.add_middleware(
                middleware,  # type: ignore[arg-type]
                **spec.extra_kwargs,
            )
            return

        # Convert the unified middleware to a FastAPI middleware class
        app.add_middleware(
            BaseHTTPMiddleware,
            dispatch=middleware,
            **spec.extra_kwargs,
        )



"""Add content hash column [aae4eed923b5].

Revision ID: aae4eed923b5
Revises: 0.84.3
Create Date: 2025-08-22 16:32:12.494008

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "aae4eed923b5"
down_revision = "0.84.3"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("artifact_version", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("content_hash", sa.TEXT(), nullable=True)
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("artifact_version", schema=None) as batch_op:
        batch_op.drop_column("content_hash")

    # ### end Alembic commands ###



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utility functions for runs."""

from typing import TYPE_CHECKING, Dict, List, Optional, Set, cast

from zenml.enums import ExecutionStatus, StackComponentType
from zenml.exceptions import IllegalOperationError
from zenml.logger import get_logger
from zenml.models import PipelineRunResponse, PipelineRunUpdate, StepRunUpdate
from zenml.orchestrators.base_orchestrator import BaseOrchestrator
from zenml.stack.stack_component import StackComponent

if TYPE_CHECKING:
    from zenml.zen_stores.base_zen_store import BaseZenStore

logger = get_logger(__name__)


def stop_run(run: PipelineRunResponse, graceful: bool = False) -> None:
    """Stop a pipeline run.

    Args:
        run: The pipeline run to stop.
        graceful: Whether to stop the run gracefully.

    Raises:
        IllegalOperationError: If the run is already stopped or being stopped.
        ValueError: If the stack is not accessible.
    """
    # Check if the stack is still accessible
    if run.stack is None:
        raise ValueError(
            "The stack that this pipeline run response was executed on "
            "is either not accessible or has been deleted."
        )

    # Check if pipeline can be stopped
    if run.status == ExecutionStatus.COMPLETED:
        raise IllegalOperationError(
            "Cannot stop a run that is already completed."
        )

    if run.status == ExecutionStatus.STOPPED:
        raise IllegalOperationError("Run is already stopped.")

    if run.status == ExecutionStatus.STOPPING:
        raise IllegalOperationError("Run is already being stopped.")

    # Check if the stack is still accessible
    orchestrator_list = run.stack.components.get(
        StackComponentType.ORCHESTRATOR, []
    )
    if len(orchestrator_list) == 0:
        raise ValueError(
            "The orchestrator that this pipeline run response was "
            "executed with is either not accessible or has been deleted."
        )

    orchestrator = cast(
        BaseOrchestrator,
        StackComponent.from_model(component_model=orchestrator_list[0]),
    )

    # Stop the run
    orchestrator.stop_run(run=run, graceful=graceful)


def refresh_run_status(
    run: PipelineRunResponse,
    include_step_updates: bool = False,
    zen_store: Optional["BaseZenStore"] = None,
) -> PipelineRunResponse:
    """Refresh the status of a pipeline run if it is initializing/running.

    Args:
        run: The pipeline run to refresh.
        include_step_updates: Flag deciding whether we should also refresh
            the status of individual steps.
        zen_store: Optional ZenStore to use for updating the run. If not provided,
            the ZenStore will be fetched from the Client. This is mainly useful
            for running inside the ZenML server.

    Returns:
        The refreshed pipeline run.

    Raises:
        ValueError: If the stack or the orchestrator of the run is deleted.
    """
    # Check if the stack still accessible
    if run.stack is None:
        raise ValueError(
            "The stack that this pipeline run response was executed on"
            "is either not accessible or has been deleted."
        )

    # Check if the orchestrator is still accessible
    orchestrator_list = run.stack.components.get(
        StackComponentType.ORCHESTRATOR, []
    )
    if len(orchestrator_list) == 0:
        raise ValueError(
            "The orchestrator that this pipeline run response was "
            "executed with is either not accessible or has been deleted."
        )

    orchestrator = cast(
        BaseOrchestrator,
        StackComponent.from_model(component_model=orchestrator_list[0]),
    )

    # Fetch the status (and optionally step statuses)
    pipeline_status, step_statuses = orchestrator.fetch_status(
        run=run, include_steps=include_step_updates
    )

    if zen_store is None:
        from zenml.client import Client

        zen_store = Client().zen_store
    else:
        zen_store = zen_store

    # Update step statuses
    if include_step_updates:
        if step_statuses:
            current_steps = run.steps
            for step_name, step_status in step_statuses.items():
                # If step is not in the run yet, skip
                if step_name not in current_steps:
                    continue

                # If the step is already finished, skip
                if current_steps[step_name].status.is_finished:
                    continue

                # Update only if the status has changed
                if step_status != current_steps[step_name].status:
                    try:
                        zen_store.update_run_step(
                            step_run_id=current_steps[step_name].id,
                            step_run_update=StepRunUpdate(status=step_status),
                        )
                    except Exception as e:
                        logger.warning(
                            f"Failed to update status for step '{step_name}': {e}"
                        )

    # Update pipeline status
    if pipeline_status is not None and pipeline_status != run.status:
        return zen_store.update_run(
            run_id=run.id,
            run_update=PipelineRunUpdate(status=pipeline_status),
        )

    return run


def build_dag(steps: Dict[str, List[str]]) -> Dict[str, Set[str]]:
    """Build DAG with downstream steps from a list of steps.

    Args:
        steps: The list of steps.

    Returns:
        The DAG with downstream steps.
    """
    dag: Dict[str, Set[str]] = {step: set() for step in steps}

    for step_name, upstream_steps in steps.items():
        for upstream_step in upstream_steps:
            dag[upstream_step].add(step_name)

    return dag


def find_all_downstream_steps(
    step_name: str, dag: Dict[str, Set[str]]
) -> Set[str]:
    """Find all downstream steps of a given step.

    Args:
        step_name: The name of the step to find the downstream steps of.
        dag: The DAG.

    Returns:
        The set of downstream steps.
    """
    downstream_steps = set()
    for downstream_step in dag[step_name]:
        downstream_steps.add(downstream_step)
        downstream_steps.update(
            find_all_downstream_steps(downstream_step, dag)
        )
    return downstream_steps



#  Copyright (c) ZenML GmbH 2026. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utility functions for the AWS integration."""

from zenml.enums import ExecutionStatus


def convert_training_job_status(status: str) -> ExecutionStatus:
    """Converts a training job status to an execution status.

    Args:
        status: The training job status.

    Raises:
        ValueError: If the training job status is unknown.

    Returns:
        The execution status.
    """
    # https://docs.aws.amazon.com/sagemaker/latest/APIReference/API_DescribeTrainingJob.html#sagemaker-DescribeTrainingJob-response-TrainingJobStatus
    if status == "InProgress":
        return ExecutionStatus.RUNNING
    elif status == "Stopping":
        return ExecutionStatus.STOPPING
    elif status == "Completed":
        return ExecutionStatus.COMPLETED
    elif status == "Failed":
        return ExecutionStatus.FAILED
    elif status == "Stopped":
        return ExecutionStatus.STOPPED
    else:
        raise ValueError(f"Unknown training job status: {status}")


def convert_processing_job_status(status: str) -> ExecutionStatus:
    """Converts a processing job status to an execution status.

    Args:
        status: The processing job status.

    Raises:
        ValueError: If the processing job status is unknown.

    Returns:
        The execution status.
    """
    # https://docs.aws.amazon.com/sagemaker/latest/APIReference/API_DescribeProcessingJob.html#sagemaker-DescribeProcessingJob-response-ProcessingJobStatus
    if status == "InProgress":
        return ExecutionStatus.RUNNING
    elif status == "Stopping":
        return ExecutionStatus.STOPPING
    elif status == "Completed":
        return ExecutionStatus.COMPLETED
    elif status == "Failed":
        return ExecutionStatus.FAILED
    else:
        raise ValueError(f"Unknown processing job status: {status}")



"""Release [0.85.0].

Revision ID: 0.85.0
Revises: 502b4fa5fa88
Create Date: 2025-09-11 14:36:30.276055

"""

# revision identifiers, used by Alembic.
revision = "0.85.0"
down_revision = "502b4fa5fa88"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""CLI functionality to interact with pipelines."""

import json
import os
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

import click

from zenml.cli import utils as cli_utils
from zenml.cli.cli import TagGroup, cli
from zenml.cli.utils import OutputFormat, fetch_snapshot, list_options
from zenml.client import Client
from zenml.console import console
from zenml.deployers.base_deployer import BaseDeployer
from zenml.enums import (
    CliCategories,
    ExecutionStatus,
    RunWaitConditionResolution,
    RunWaitConditionStatus,
)
from zenml.logger import get_logger
from zenml.models import (
    PipelineBuildBase,
    PipelineBuildFilter,
    PipelineFilter,
    PipelineRunFilter,
    PipelineRunUpdate,
    PipelineSnapshotFilter,
    RunWaitConditionFilter,
    ScheduleFilter,
)
from zenml.pipelines.pipeline_definition import Pipeline
from zenml.utils import run_utils, source_utils, uuid_utils
from zenml.utils.dashboard_utils import get_deployment_url
from zenml.utils.yaml_utils import write_yaml

logger = get_logger(__name__)


def _import_pipeline(source: str) -> Pipeline:
    """Import a pipeline.

    Args:
        source: The pipeline source.

    Returns:
        The pipeline.
    """
    try:
        pipeline_instance = source_utils.load(source)
    except ModuleNotFoundError as e:
        source_root = source_utils.get_source_root()
        cli_utils.error(
            f"Unable to import module `{e.name}`. Make sure the source path is "
            f"relative to your source root `{source_root}`."
        )
    except AttributeError as e:
        cli_utils.error("Unable to load attribute from module: " + str(e))

    if not isinstance(pipeline_instance, Pipeline):
        cli_utils.error(
            f"The given source path `{source}` does not resolve to a pipeline "
            "object."
        )

    return pipeline_instance


@cli.group(cls=TagGroup, tag=CliCategories.MANAGEMENT_TOOLS)
def pipeline() -> None:
    """Interact with pipelines, runs and schedules."""


@pipeline.command(
    "register",
    help="Register a pipeline instance. The SOURCE argument needs to be an "
    "importable source path resolving to a ZenML pipeline instance, e.g. "
    "`my_module.my_pipeline_instance`.",
)
@click.argument("source")
@click.option(
    "--parameters",
    "-p",
    "parameters_path",
    type=click.Path(exists=True, dir_okay=False),
    required=False,
    help="Path to JSON file containing parameters for the pipeline function.",
)
def register_pipeline(
    source: str, parameters_path: Optional[str] = None
) -> None:
    """Register a pipeline.

    Args:
        source: Importable source resolving to a pipeline instance.
        parameters_path: Path to pipeline parameters file.
    """
    if "." not in source:
        cli_utils.error(
            f"The given source path `{source}` is invalid. Make sure it looks "
            "like `some.module.name_of_pipeline_instance_variable` and "
            "resolves to a pipeline object."
        )

    if not Client().root:
        cli_utils.warning(
            "You're running the `zenml pipeline register` command without a "
            "ZenML repository. Your current working directory will be used "
            "as the source root relative to which the `source` argument is "
            "expected. To silence this warning, run `zenml init` at your "
            "source code root."
        )

    pipeline_instance = _import_pipeline(source=source)

    parameters: Dict[str, Any] = {}
    if parameters_path:
        with open(parameters_path, "r") as f:
            parameters = json.load(f)

    try:
        pipeline_instance.prepare(**parameters)
    except ValueError:
        cli_utils.error(
            "Pipeline preparation failed. This is most likely due to your "
            "pipeline entrypoint function requiring arguments that were not "
            "provided. Please provide a JSON file with the parameters for "
            f"your pipeline like this: `zenml pipeline register {source} "
            "--parameters=<PATH_TO_JSON>`."
        )

    pipeline_instance.register()


@pipeline.command(
    "build",
    help="Build Docker images for a pipeline. The SOURCE argument needs to be "
    " an importable source path resolving to a ZenML pipeline instance, e.g. "
    "`my_module.my_pipeline_instance`.",
)
@click.argument("source")
@click.option(
    "--config",
    "-c",
    "config_path",
    type=click.Path(exists=True, dir_okay=False),
    required=False,
    help="Path to configuration file for the build.",
)
@click.option(
    "--stack",
    "-s",
    "stack_name_or_id",
    type=str,
    required=False,
    help="Name or ID of the stack to use for the build.",
)
@click.option(
    "--output",
    "-o",
    "output_path",
    type=click.Path(exists=False, dir_okay=False),
    required=False,
    help="Output path for the build information.",
)
def build_pipeline(
    source: str,
    config_path: Optional[str] = None,
    stack_name_or_id: Optional[str] = None,
    output_path: Optional[str] = None,
) -> None:
    """Build Docker images for a pipeline.

    Args:
        source: Importable source resolving to a pipeline instance.
        config_path: Path to pipeline configuration file.
        stack_name_or_id: Name or ID of the stack for which the images should
            be built.
        output_path: Optional file path to write the output to.
    """
    if not Client().root:
        cli_utils.warning(
            "You're running the `zenml pipeline build` command without a "
            "ZenML repository. Your current working directory will be used "
            "as the source root relative to which the registered step classes "
            "will be resolved. To silence this warning, run `zenml init` at "
            "your source code root."
        )

    with cli_utils.temporary_active_stack(stack_name_or_id=stack_name_or_id):
        pipeline_instance = _import_pipeline(source=source)

        pipeline_instance = pipeline_instance.with_options(
            config_path=config_path
        )
        build = pipeline_instance.build()

    if build:
        cli_utils.declare(f"Created pipeline build `{build.id}`.")

        if output_path:
            cli_utils.declare(
                f"Writing pipeline build output to `{output_path}`."
            )
            write_yaml(output_path, build.to_yaml())
    else:
        cli_utils.declare("No docker builds required.")


@pipeline.command(
    "run",
    help="Run a pipeline. The SOURCE argument needs to be an "
    "importable source path resolving to a ZenML pipeline instance, e.g. "
    "`my_module.my_pipeline_instance`.",
)
@click.argument("source")
@click.option(
    "--config",
    "-c",
    "config_path",
    type=click.Path(exists=True, dir_okay=False),
    required=False,
    help="Path to configuration file for the run.",
)
@click.option(
    "--stack",
    "-s",
    "stack_name_or_id",
    type=str,
    required=False,
    help="Name or ID of the stack to run on.",
)
@click.option(
    "--build",
    "-b",
    "build_path_or_id",
    type=str,
    required=False,
    help="ID or path of the build to use.",
)
@click.option(
    "--prevent-build-reuse",
    is_flag=True,
    default=False,
    required=False,
    help="Prevent automatic build reusing.",
)
def run_pipeline(
    source: str,
    config_path: Optional[str] = None,
    stack_name_or_id: Optional[str] = None,
    build_path_or_id: Optional[str] = None,
    prevent_build_reuse: bool = False,
) -> None:
    """Run a pipeline.

    Args:
        source: Importable source resolving to a pipeline instance.
        config_path: Path to pipeline configuration file.
        stack_name_or_id: Name or ID of the stack on which the pipeline should
            run.
        build_path_or_id: ID of file path of the build to use for the pipeline
            run.
        prevent_build_reuse: If True, prevents automatic reusing of previous
            builds.
    """
    if not Client().root:
        cli_utils.warning(
            "You're running the `zenml pipeline run` command without a "
            "ZenML repository. Your current working directory will be used "
            "as the source root relative to which the registered step classes "
            "will be resolved. To silence this warning, run `zenml init` at "
            "your source code root."
        )

    with cli_utils.temporary_active_stack(stack_name_or_id=stack_name_or_id):
        pipeline_instance = _import_pipeline(source=source)

        build: Union[str, PipelineBuildBase, None] = None
        if build_path_or_id:
            if uuid_utils.is_valid_uuid(build_path_or_id):
                build = build_path_or_id
            elif os.path.exists(build_path_or_id):
                build = PipelineBuildBase.from_yaml(build_path_or_id)
            else:
                cli_utils.error(
                    f"The specified build {build_path_or_id} is not a valid UUID "
                    "or file path."
                )

        pipeline_instance = pipeline_instance.with_options(
            config_path=config_path,
            build=build,
            prevent_build_reuse=prevent_build_reuse,
        )
        pipeline_instance()


@pipeline.command(
    "deploy",
    help="Deploy a pipeline. The SOURCE argument needs to be an "
    "importable source path resolving to a ZenML pipeline instance, e.g. "
    "`my_module.my_pipeline_instance`.",
)
@click.argument("source")
@click.option(
    "--name",
    "-n",
    "deployment_name",
    type=str,
    required=False,
    help="The name of the deployment resulted from deploying the pipeline. If "
    "not provided, the name of the pipeline will be used. If an existing "
    "deployment with the same name already exists, an error will be raised, "
    "unless the --update or --overtake flag is used.",
)
@click.option(
    "--config",
    "-c",
    "config_path",
    type=click.Path(exists=True, dir_okay=False),
    required=False,
    help="Path to configuration file for the deployment.",
)
@click.option(
    "--stack",
    "-s",
    "stack_name_or_id",
    type=str,
    required=False,
    help="Name or ID of the stack to deploy on.",
)
@click.option(
    "--build",
    "-b",
    "build_path_or_id",
    type=str,
    required=False,
    help="ID or path of the build to use.",
)
@click.option(
    "--prevent-build-reuse",
    is_flag=True,
    default=False,
    required=False,
    help="Prevent automatic build reusing.",
)
@click.option(
    "--update",
    "-u",
    "update",
    is_flag=True,
    default=False,
    required=False,
    help="Update the deployment with the same name if it already exists.",
)
@click.option(
    "--overtake",
    "-o",
    "overtake",
    is_flag=True,
    default=False,
    required=False,
    help="Update the deployment with the same name if it already "
    "exists, even if it is owned by a different user.",
)
@click.option(
    "--attach",
    "-a",
    "attach",
    is_flag=True,
    default=False,
    required=False,
    help="Attach to the deployment logs.",
)
@click.option(
    "--timeout",
    "-t",
    "timeout",
    type=int,
    required=False,
    default=None,
    help="Maximum time in seconds to wait for the pipeline to be deployed.",
)
def deploy_pipeline(
    source: str,
    deployment_name: Optional[str] = None,
    config_path: Optional[str] = None,
    stack_name_or_id: Optional[str] = None,
    build_path_or_id: Optional[str] = None,
    prevent_build_reuse: bool = False,
    update: bool = False,
    overtake: bool = False,
    attach: bool = False,
    timeout: Optional[int] = None,
) -> None:
    """Deploy a pipeline for online inference.

    Args:
        source: Importable source resolving to a pipeline instance.
        deployment_name: Name of the deployment used to deploy the pipeline on.
        config_path: Path to pipeline configuration file.
        stack_name_or_id: Name or ID of the stack on which the pipeline should
            be deployed.
        build_path_or_id: ID of file path of the build to use for the pipeline
            deployment.
        prevent_build_reuse: If True, prevents automatic reusing of previous
            builds.
        update: If True, update the deployment with the same name if it
            already exists.
        overtake: If True, update the deployment with the same name if
            it already exists, even if it is owned by a different user.
        attach: If True, attach to the deployment logs.
        timeout: The maximum time in seconds to wait for the pipeline to be
            deployed.
    """
    if not Client().root:
        cli_utils.warning(
            "You're running the `zenml pipeline deploy` command without a "
            "ZenML repository. Your current working directory will be used "
            "as the source root relative to which the registered step classes "
            "will be resolved. To silence this warning, run `zenml init` at "
            "your source code root."
        )

    with cli_utils.temporary_active_stack(stack_name_or_id=stack_name_or_id):
        pipeline_instance = _import_pipeline(source=source)

        build: Union[str, PipelineBuildBase, None] = None
        if build_path_or_id:
            if uuid_utils.is_valid_uuid(build_path_or_id):
                build = build_path_or_id
            elif os.path.exists(build_path_or_id):
                build = PipelineBuildBase.from_yaml(build_path_or_id)
            else:
                cli_utils.error(
                    f"The specified build {build_path_or_id} is not a valid UUID "
                    "or file path."
                )

        pipeline_instance = pipeline_instance.with_options(
            config_path=config_path,
            build=build,
            prevent_build_reuse=prevent_build_reuse,
        )
        if not deployment_name:
            deployment_name = pipeline_instance.name
        client = Client()
        try:
            deployment = client.get_deployment(deployment_name)
        except KeyError:
            pass
        else:
            if (
                deployment.user
                and deployment.user.id != client.active_user.id
                and not overtake
            ):
                confirmation = cli_utils.confirmation(
                    f"Deployment with name '{deployment_name}' already exists "
                    f"and is owned by a different user '{deployment.user.name}'."
                    "\nDo you want to continue and update the existing deployment "
                    "(hint: use the --overtake flag to skip this check) ?"
                )
                if not confirmation:
                    cli_utils.declare("Deployment canceled.")
                    return
            elif not update and not overtake:
                confirmation = cli_utils.confirmation(
                    f"Deployment with name '{deployment_name}' already exists.\n"
                    "Do you want to continue and update the existing "
                    "deployment "
                    "(hint: use the --update flag to skip this check) ?"
                )
                if not confirmation:
                    cli_utils.declare("Deployment canceled.")
                    return

        deployment = pipeline_instance.deploy(
            deployment_name=deployment_name, timeout=timeout
        )

        cli_utils.pretty_print_deployment(deployment, show_secret=False)

        dashboard_url = get_deployment_url(deployment)
        if dashboard_url:
            cli_utils.declare(
                f"\nView in the ZenML UI: [link]{dashboard_url}[/link]",
                no_wrap=True,
            )

        if attach:
            deployer = BaseDeployer.get_active_deployer()
            for log in deployer.get_deployment_logs(
                deployment_name_or_id=deployment.id,
                follow=True,
            ):
                print(log)


@pipeline.command(
    "create-run-template",
    help="Create a run template for a pipeline. The SOURCE argument needs to "
    "be an importable source path resolving to a ZenML pipeline instance, e.g. "
    "`my_module.my_pipeline_instance`.",
)
@click.argument("source")
@click.option(
    "--name",
    "-n",
    type=str,
    required=True,
    help="Name for the template",
)
@click.option(
    "--config",
    "-c",
    "config_path",
    type=click.Path(exists=True, dir_okay=False),
    required=False,
    help="Path to configuration file for the build.",
)
@click.option(
    "--stack",
    "-s",
    "stack_name_or_id",
    type=str,
    required=False,
    help="Name or ID of the stack to use for the build.",
)
def create_run_template(
    source: str,
    name: str,
    config_path: Optional[str] = None,
    stack_name_or_id: Optional[str] = None,
) -> None:
    """DEPRECATED: Create a run template for a pipeline.

    Args:
        source: Importable source resolving to a pipeline instance.
        name: Name of the run template.
        config_path: Path to pipeline configuration file.
        stack_name_or_id: Name or ID of the stack for which the template should
            be created.
    """
    cli_utils.warning(
        "The `zenml pipeline create-run-template` command is deprecated and "
        "will be removed in a future version. Please use `zenml pipeline "
        "snapshot create` instead."
    )
    if not Client().root:
        cli_utils.warning(
            "You're running the `zenml pipeline create-run-template` command "
            "without a ZenML repository. Your current working directory will "
            "be used as the source root relative to which the registered step "
            "classes will be resolved. To silence this warning, run `zenml "
            "init` at your source code root."
        )

    with cli_utils.temporary_active_stack(stack_name_or_id=stack_name_or_id):
        pipeline_instance = _import_pipeline(source=source)

        pipeline_instance = pipeline_instance.with_options(
            config_path=config_path
        )
        template = pipeline_instance.create_run_template(name=name)

    cli_utils.declare(f"Created run template `{template.id}`.")


@pipeline.command("list", help="List all registered pipelines.")
@list_options(
    PipelineFilter,
    default_columns=["id", "name", "user", "latest_run_status", "created"],
)
def list_pipelines(
    columns: str, output_format: OutputFormat, **kwargs: Any
) -> None:
    """List all registered pipelines.

    Args:
        columns: Columns to display in output.
        output_format: Format for output (table/json/yaml/csv/tsv).
        **kwargs: Keyword arguments to filter pipelines.
    """
    client = Client()
    with console.status("Listing pipelines...\n"):
        pipelines = client.list_pipelines(**kwargs)

    cli_utils.print_page(
        pipelines,
        columns,
        output_format,
        empty_message="No pipelines found for this filter.",
    )


@pipeline.command("delete")
@click.argument("pipeline_name_or_id", type=str, required=True)
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    help="Don't ask for confirmation.",
)
def delete_pipeline(
    pipeline_name_or_id: str,
    yes: bool = False,
) -> None:
    """Delete a pipeline.

    Args:
        pipeline_name_or_id: The name or ID of the pipeline to delete.
        yes: If set, don't ask for confirmation.
    """
    if not yes:
        confirmation = cli_utils.confirmation(
            f"Are you sure you want to delete pipeline "
            f"`{pipeline_name_or_id}`? This will delete all "
            "runs and snapshots of this pipeline."
        )
        if not confirmation:
            cli_utils.declare("Pipeline deletion canceled.")
            return

    try:
        Client().delete_pipeline(
            name_id_or_prefix=pipeline_name_or_id,
        )
    except KeyError as e:
        cli_utils.exception(e)
    else:
        cli_utils.declare(f"Deleted pipeline `{pipeline_name_or_id}`.")


@pipeline.group()
def schedule() -> None:
    """Commands for pipeline run schedules."""


@schedule.command("list", help="List all pipeline schedules.")
@list_options(
    ScheduleFilter,
    default_columns=[
        "id",
        "name",
        "pipeline",
        "cron_expression",
        "active",
        "is_archived",
    ],
)
def list_schedules(
    columns: str, output_format: OutputFormat, **kwargs: Any
) -> None:
    """List all pipeline schedules.

    Args:
        columns: Columns to display in output.
        output_format: Format for output (table/json/yaml/csv/tsv).
        **kwargs: Keyword arguments to filter schedules.
    """
    schedules = Client().list_schedules(**kwargs)
    cli_utils.print_page(
        schedules,
        columns,
        output_format,
        row_formatter=cli_utils.generate_schedule_row,
        empty_message="No schedules found for this filter.",
    )


@schedule.command("update", help="Update a pipeline schedule.")
@click.argument("schedule_name_or_id", type=str, required=True)
@click.option(
    "--cron-expression",
    "-c",
    type=str,
    required=False,
    help="The cron expression to update the schedule with.",
)
def update_schedule(
    schedule_name_or_id: str,
    cron_expression: str | None = None,
) -> None:
    """Update a pipeline schedule.

    Args:
        schedule_name_or_id: The name or ID of the schedule to update.
        cron_expression: The cron expression to update the schedule with.
    """
    options = [
        cron_expression,
    ]

    if not any(options):
        cli_utils.declare("No schedule update requested.")
        return

    try:
        Client().update_schedule(
            name_id_or_prefix=schedule_name_or_id,
            cron_expression=cron_expression,
        )
    except Exception as e:
        cli_utils.exception(e)
    else:
        cli_utils.declare(f"Updated schedule '{schedule_name_or_id}'.")


@schedule.command("activate", help="Activate a pipeline schedule.")
@click.argument("schedule_name_or_id", type=str, required=True)
def activate_schedule(schedule_name_or_id: str) -> None:
    """Activate a pipeline schedule.

    Args:
        schedule_name_or_id: The name or ID of the schedule to delete.
    """
    try:
        Client().update_schedule(
            name_id_or_prefix=schedule_name_or_id, active=True
        )
    except KeyError as e:
        cli_utils.exception(e)
    else:
        cli_utils.declare(f"Activated schedule '{schedule_name_or_id}'.")


@schedule.command("deactivate", help="Deactivate a pipeline schedule.")
@click.argument("schedule_name_or_id", type=str, required=True)
def deactivate_schedule(schedule_name_or_id: str) -> None:
    """Activate a pipeline schedule.

    Args:
        schedule_name_or_id: The name or ID of the schedule to delete.
    """
    try:
        Client().update_schedule(
            name_id_or_prefix=schedule_name_or_id, active=False
        )
    except KeyError as e:
        cli_utils.exception(e)
    else:
        cli_utils.declare(f"Deactivated schedule '{schedule_name_or_id}'.")


@schedule.command("delete", help="Delete a pipeline schedule.")
@click.argument("schedule_name_or_id", type=str, required=True)
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    help="Don't ask for confirmation.",
)
@click.option("--hard", "-h", is_flag=True, help="Hard delete the schedule")
def delete_schedule(
    schedule_name_or_id: str, yes: bool = False, hard: bool = False
) -> None:
    """Delete a pipeline schedule.

    Args:
        schedule_name_or_id: The name or ID of the schedule to delete.
        yes: If set, don't ask for confirmation.
        hard: If set will trigger hard deletion of the schedule instead of archiving.
    """
    if not yes:
        if not hard:
            confirmation = cli_utils.confirmation(
                f"Are you sure you want to delete schedule "
                f"`{schedule_name_or_id}`?"
            )
        else:
            confirmation = cli_utils.confirmation(
                f"Are you sure you want to hard delete schedule "
                f"`{schedule_name_or_id}`? Any historical data or references "
                f"to this schedule will be lost."
            )
        if not confirmation:
            cli_utils.declare("Schedule deletion canceled.")
            return

    try:
        Client().delete_schedule(
            name_id_or_prefix=schedule_name_or_id, soft=not hard
        )
    except KeyError as e:
        cli_utils.exception(e)
    else:
        cli_utils.declare(f"Deleted schedule '{schedule_name_or_id}'.")


@pipeline.group()
def runs() -> None:
    """Commands for pipeline runs."""


@runs.command("list", help="List all registered pipeline runs.")
@list_options(
    PipelineRunFilter,
    default_columns=[
        "id",
        "index",
        "run_name",
        "pipeline",
        "status",
        "stack",
        "owner",
    ],
)
def list_pipeline_runs(
    columns: str, output_format: OutputFormat, **kwargs: Any
) -> None:
    """List all registered pipeline runs for the filter.

    Args:
        columns: Columns to display in output.
        output_format: Format for output (table/json/yaml/csv/tsv).
        **kwargs: Keyword arguments to filter pipeline runs.
    """
    client = Client()
    try:
        with console.status("Listing pipeline runs...\n"):
            pipeline_runs = client.list_pipeline_runs(**kwargs)
    except KeyError as err:
        cli_utils.exception(err)
        return

    cli_utils.print_page(
        pipeline_runs,
        columns,
        output_format,
        cli_utils.generate_pipeline_run_row,
        empty_message="No pipeline runs found for this filter.",
    )


@runs.group(name="wait-conditions")
def run_wait_conditions() -> None:
    """Commands for run wait conditions."""


@run_wait_conditions.command("list")
@list_options(
    RunWaitConditionFilter,
    default_columns=[
        "id",
        "name",
        "type",
        "status",
        "created",
    ],
)
def list_run_wait_conditions(
    columns: str,
    output_format: OutputFormat,
    **kwargs: Any,
) -> None:
    """List run wait conditions.

    Args:
        columns: Columns to display.
        output_format: Format for output.
        **kwargs: Additional wait condition filters.
    """
    client = Client()
    try:
        with console.status("Listing wait conditions...\n"):
            conditions = client.list_run_wait_conditions(**kwargs)
    except KeyError as err:
        cli_utils.exception(err)

    cli_utils.print_page(
        conditions,
        columns,
        output_format,
        empty_message="No wait conditions found.",
    )


@run_wait_conditions.command("resolve")
@click.argument("wait_condition_id", type=str, required=False)
@click.option(
    "--resolution",
    type=click.Choice([r.value for r in RunWaitConditionResolution]),
    required=False,
    help="Condition resolution.",
)
@click.option(
    "--result",
    type=str,
    required=False,
    help="Optional JSON result for the condition.",
)
@click.option(
    "--interactive",
    "-i",
    is_flag=True,
    default=False,
    help="Interactively review and resolve pending wait conditions.",
)
@click.option(
    "--run",
    "-r",
    "run_name_or_id",
    type=str,
    required=False,
    help="Optional run name/ID to scope interactive resolution.",
)
def resolve_wait_condition(
    wait_condition_id: Optional[str] = None,
    resolution: Optional[str] = None,
    result: Optional[str] = None,
    interactive: bool = False,
    run_name_or_id: Optional[str] = None,
) -> None:
    """Resolve run wait conditions.

    Args:
        wait_condition_id: Optional wait condition ID to resolve or review.
        resolution: Optional resolution semantic value.
        result: Optional JSON value with result.
        interactive: Whether to interactively review pending wait conditions.
        run_name_or_id: Optional run name or ID to scope interactive resolution.
    """
    if interactive:
        if resolution is not None or result is not None:
            cli_utils.error(
                "`--interactive` cannot be used with `--resolution` or "
                "`--result`."
            )
        _interactive_resolve_wait_conditions(
            wait_condition_id=wait_condition_id,
            run_name_or_id=run_name_or_id,
        )
        return

    if wait_condition_id is None:
        cli_utils.error(
            "Argument `WAIT_CONDITION_ID` is required in non-interactive mode."
        )
    if resolution is None:
        cli_utils.error(
            "Option `--resolution` is required in non-interactive mode."
        )

    resolution_enum = RunWaitConditionResolution(resolution)
    if (
        resolution_enum == RunWaitConditionResolution.ABORT
        and result is not None
    ):
        cli_utils.error("`--result` cannot be used with `--resolution abort`.")

    result_value: Optional[Any] = None
    if result:
        try:
            result_value = json.loads(result)
        except json.JSONDecodeError as e:
            cli_utils.error(f"Invalid JSON for `--result`: {e}")

    try:
        condition_id = UUID(wait_condition_id)
    except ValueError:
        cli_utils.error(f"Invalid wait condition ID: `{wait_condition_id}`.")

    try:
        Client().resolve_run_wait_condition(
            run_wait_condition_id=condition_id,
            resolution=resolution_enum,
            result=result_value,
        )
    except KeyError as err:
        cli_utils.exception(err)

    cli_utils.declare(
        f"Resolved wait condition `{wait_condition_id}` with resolution "
        f"`{resolution_enum.value}`."
    )


def _interactive_resolve_wait_conditions(
    wait_condition_id: Optional[str] = None,
    run_name_or_id: Optional[str] = None,
) -> None:
    """Interactively resolve pending wait conditions.

    Args:
        wait_condition_id: Optional wait condition ID to review.
        run_name_or_id: Optional run name or ID to narrow the resolution scope.
    """
    client = Client()
    skipped_condition_ids: set[UUID] = set()

    def _resolve_single_condition_interactively(
        condition: Any,
        idx: int,
        total: int,
    ) -> str:
        """Interactively handle a single wait condition.

        Args:
            condition: The wait condition to resolve.
            idx: The 1-based position of the condition in the current batch.
            total: The total number of conditions in the current batch.

        Returns:
            One of `resolved`, `skipped`, or `quit`.
        """
        run_name = condition.run.name
        run_id = condition.run.id

        click.echo("")
        click.echo(f"[{idx}/{total}] Run `{run_name}` ({run_id})")
        click.echo(f"Condition `{condition.id}` name={condition.name}")
        click.echo(
            f"Type: {condition.type.value} | Status: {condition.status}"
        )

        if condition.question:
            click.echo("\nQuestion:")
            click.echo(condition.question)

        expected_schema = condition.data_schema
        if expected_schema is not None:
            click.echo("\nExpected JSON schema:")
            click.echo(json.dumps(expected_schema, indent=2, sort_keys=True))

        while True:
            action = (
                click.prompt(
                    "\nAction [c=continue, a=abort, s=skip, q=quit]",
                    type=str,
                    default="c",
                )
                .strip()
                .lower()
            )
            if action not in {"c", "a", "s", "q"}:
                click.echo("invalid action", err=True)
                continue

            if action == "q":
                return "quit"
            if action == "s":
                return "skipped"

            resolution = (
                RunWaitConditionResolution.CONTINUE
                if action == "c"
                else RunWaitConditionResolution.ABORT
            )

            result: Optional[Any] = None
            if (
                resolution == RunWaitConditionResolution.CONTINUE
                and expected_schema is not None
            ):
                while True:
                    raw_value = click.prompt(
                        "JSON value for `result` (empty for null)",
                        type=str,
                        default="",
                        show_default=False,
                    )
                    if raw_value == "":
                        result = None
                        break
                    try:
                        result = json.loads(raw_value)
                        break
                    except json.JSONDecodeError as e:
                        click.echo(f"Invalid JSON value: {e}", err=True)

            try:
                client.resolve_run_wait_condition(
                    run_wait_condition_id=condition.id,
                    resolution=resolution,
                    result=result,
                )
                cli_utils.declare(
                    f"Resolved wait condition `{condition.id}` with resolution "
                    f"`{resolution.value}`."
                )
                return "resolved"
            except Exception as e:
                click.echo(f"Failed to resolve wait condition: {e}", err=True)

    if wait_condition_id is not None:
        try:
            condition = client.zen_store.get_run_wait_condition(
                UUID(wait_condition_id), hydrate=True
            )
        except ValueError:
            cli_utils.error(
                f"Invalid wait condition ID: `{wait_condition_id}`."
            )
        if condition.status != RunWaitConditionStatus.PENDING:
            cli_utils.error(
                f"Wait condition `{wait_condition_id}` is currently "
                f"`{condition.status.value}`, not `pending`."
            )
        _resolve_single_condition_interactively(
            condition=condition,
            idx=1,
            total=1,
        )
        return
    else:
        while True:
            page = 1
            conditions = []
            while True:
                page_result = client.list_run_wait_conditions(
                    project=client.active_project.id,
                    pipeline_run=run_name_or_id,
                    status=RunWaitConditionStatus.PENDING.value,
                    sort_by="asc:created",
                    page=page,
                    size=20,
                    hydrate=True,
                )
                conditions = [
                    condition
                    for condition in page_result.items
                    if condition.id not in skipped_condition_ids
                ]
                if conditions or page >= page_result.total_pages:
                    break
                page += 1

            if not conditions:
                if skipped_condition_ids:
                    cli_utils.declare(
                        "No additional pending wait conditions found."
                    )
                elif run_name_or_id:
                    cli_utils.declare(
                        f"No pending wait conditions found for run "
                        f"`{run_name_or_id}`."
                    )
                else:
                    cli_utils.declare("No pending wait conditions found.")
                return

            for idx, condition in enumerate(conditions, start=1):
                result = _resolve_single_condition_interactively(
                    condition=condition,
                    idx=idx,
                    total=len(conditions),
                )
                if result == "quit":
                    return
                if result == "skipped":
                    skipped_condition_ids.add(condition.id)


@runs.command("stop")
@click.argument("run_name_or_id", type=str, required=True)
@click.option(
    "--graceful",
    "-g",
    is_flag=True,
    default=False,
    help="Use graceful shutdown (default is False).",
)
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help="Don't ask for confirmation.",
)
def stop_pipeline_run(
    run_name_or_id: str,
    graceful: bool = False,
    yes: bool = False,
) -> None:
    """Stop a running pipeline.

    Args:
        run_name_or_id: The name or ID of the pipeline run to stop.
        graceful: If True, uses graceful shutdown. If False, forces immediate termination.
        yes: If set, don't ask for confirmation.
    """
    # Ask for confirmation to stop run.
    if not yes:
        action = "gracefully stop" if graceful else "force stop"
        confirmation = cli_utils.confirmation(
            f"Are you sure you want to {action} pipeline run `{run_name_or_id}`?"
        )
        if not confirmation:
            cli_utils.declare("Not stopping the pipeline run.")
            return

    # Stop run.
    try:
        run = Client().get_pipeline_run(name_id_or_prefix=run_name_or_id)
        run_utils.stop_run(run=run, graceful=graceful)
        action = "Gracefully stopped" if graceful else "Force stopped"
        cli_utils.declare(f"{action} pipeline run '{run.name}'.")
    except NotImplementedError:
        cli_utils.error(
            "The orchestrator used for this pipeline run does not support "
            f"{'gracefully' if graceful else 'forcefully'} stopping runs."
        )
    except Exception as e:
        cli_utils.error(f"Failed to stop pipeline run: {e}")


@runs.command("resume")
@click.argument("run_name_or_id", type=str, required=True)
def resume_pipeline_run(run_name_or_id: str) -> None:
    """Resume a pipeline run.

    Args:
        run_name_or_id: The name or ID of the pipeline run to resume.
    """
    try:
        client = Client()
        run = client.get_pipeline_run(
            name_id_or_prefix=run_name_or_id,
            hydrate=True,
        )

        if run.status != ExecutionStatus.PAUSED:
            cli_utils.error("Cannot resume a run that is not paused.")

        if run.active_wait_condition:
            cli_utils.error(
                "Cannot resume a run that has an active wait condition. "
                "Please resolve it before resuming the run:\n"
                "`zenml pipeline runs wait-conditions resolve "
                f"--run {run.id} --interactive`"
            )

        snapshot = run.snapshot
        if snapshot is None:
            cli_utils.error(
                "Unable to resume run because the snapshot is missing."
            )

        if not snapshot.stack:
            cli_utils.error(
                "Unable to resume run because the stack is missing."
            )

        client.zen_store.update_run(
            run_id=run.id,
            run_update=PipelineRunUpdate(
                status=ExecutionStatus.RESUMING,
                status_reason="Manual resume requested by user.",
            ),
        )
        try:
            with cli_utils.temporary_active_stack(
                stack_name_or_id=snapshot.stack.id
            ) as stack:
                stack.orchestrator.resume_run(
                    snapshot=snapshot, run=run, stack=stack
                )
        except Exception:
            client.zen_store.update_run(
                run_id=run.id,
                run_update=PipelineRunUpdate(
                    status=ExecutionStatus.PAUSED,
                    status_reason="Manual resume failed.",
                ),
            )
            raise
    except Exception as e:
        cli_utils.error(f"Failed to resume pipeline run: {e}")


@runs.command("retry")
@click.argument("run_name_or_id", type=str, required=True)
def retry_pipeline_run(run_name_or_id: str) -> None:
    """Retry a pipeline run by restarting its orchestrator run.

    Args:
        run_name_or_id: The name or ID of the pipeline run to retry.

    Raises:
        RuntimeError: If retrying the run fails.
    """
    try:
        client = Client()
        run = client.get_pipeline_run(
            name_id_or_prefix=run_name_or_id,
            hydrate=True,
        )

        if run.status != ExecutionStatus.FAILED:
            cli_utils.error("Cannot retry a run that is not failed.")

        snapshot = run.snapshot
        if snapshot is None:
            cli_utils.error(
                "Unable to retry run because the snapshot is missing."
            )

        if not snapshot.stack:
            cli_utils.error(
                "Unable to retry run because the stack is missing."
            )

        client.zen_store.update_run(
            run_id=run.id,
            run_update=PipelineRunUpdate(
                status=ExecutionStatus.RESUMING,
                status_reason="Manual retry requested by user.",
            ),
        )

        try:
            with cli_utils.temporary_active_stack(
                stack_name_or_id=snapshot.stack.id
            ) as stack:
                stack.orchestrator.resume_run(
                    snapshot=snapshot, run=run, stack=stack
                )
        except Exception as e:
            try:
                client.zen_store.update_run(
                    run_id=run.id,
                    run_update=PipelineRunUpdate(
                        status=ExecutionStatus.FAILED,
                        status_reason="Retry submission failed.",
                    ),
                )
            except Exception as rollback_error:
                raise RuntimeError(
                    "Retry submission failed and the run status could not be "
                    f"rolled back: {rollback_error}"
                ) from e

            raise
    except Exception as e:
        cli_utils.error(f"Failed to retry pipeline run: {e}")


@runs.command("delete")
@click.argument("run_name_or_id", type=str, required=True)
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    help="Don't ask for confirmation.",
)
def delete_pipeline_run(
    run_name_or_id: str,
    yes: bool = False,
) -> None:
    """Delete a pipeline run.

    Args:
        run_name_or_id: The name or ID of the pipeline run to delete.
        yes: If set, don't ask for confirmation.
    """
    # Ask for confirmation to delete run.
    if not yes:
        confirmation = cli_utils.confirmation(
            f"Are you sure you want to delete pipeline run `{run_name_or_id}`?"
        )
        if not confirmation:
            cli_utils.declare("Pipeline run deletion canceled.")
            return

    # Delete run.
    try:
        Client().delete_pipeline_run(
            name_id_or_prefix=run_name_or_id,
        )
    except KeyError as e:
        cli_utils.exception(e)
    else:
        cli_utils.declare(f"Deleted pipeline run '{run_name_or_id}'.")


@runs.command("refresh")
@click.argument("run_name_or_id", type=str, required=True)
@click.option(
    "--include-steps",
    is_flag=True,
    default=False,
    help="Also refresh the status of individual steps.",
)
def refresh_pipeline_run(
    run_name_or_id: str, include_steps: bool = False
) -> None:
    """Refresh the status of a pipeline run.

    Args:
        run_name_or_id: The name or ID of the pipeline run to refresh.
        include_steps: If True, also refresh the status of individual steps.
    """
    try:
        # Fetch and update the run
        run = Client().get_pipeline_run(name_id_or_prefix=run_name_or_id)
        run_utils.refresh_run_status(
            run=run, include_step_updates=include_steps
        )

    except KeyError as e:
        cli_utils.exception(e)
    else:
        cli_utils.declare(
            f"Refreshed the status of pipeline run '{run.name}'."
        )


@pipeline.group()
def builds() -> None:
    """Commands for pipeline builds."""


@builds.command("list", help="List all pipeline builds.")
@list_options(
    PipelineBuildFilter,
    default_columns=[
        "id",
        "pipeline",
        "stack",
        "zenml_version",
        "python_version",
        "checksum",
        "is_local",
    ],
)
def list_pipeline_builds(
    columns: str, output_format: OutputFormat, **kwargs: Any
) -> None:
    """List all pipeline builds for the filter.

    Args:
        columns: Columns to display in output.
        output_format: Format for output (table/json/yaml/csv/tsv).
        **kwargs: Keyword arguments to filter pipeline builds.
    """
    client = Client()
    try:
        with console.status("Listing pipeline builds...\n"):
            pipeline_builds = client.list_builds(hydrate=True, **kwargs)
    except KeyError as err:
        cli_utils.exception(err)
        return

    cli_utils.print_page(
        pipeline_builds,
        columns,
        output_format,
        empty_message="No pipeline builds found for this filter.",
    )


@builds.command("delete")
@click.argument("build_id", type=str, required=True)
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    help="Don't ask for confirmation.",
)
def delete_pipeline_build(
    build_id: str,
    yes: bool = False,
) -> None:
    """Delete a pipeline build.

    Args:
        build_id: The ID of the pipeline build to delete.
        yes: If set, don't ask for confirmation.
    """
    if not yes:
        confirmation = cli_utils.confirmation(
            f"Are you sure you want to delete pipeline build `{build_id}`?"
        )
        if not confirmation:
            cli_utils.declare("Pipeline build deletion canceled.")
            return

    try:
        Client().delete_build(build_id)
    except KeyError as e:
        cli_utils.exception(e)
    else:
        cli_utils.declare(f"Deleted pipeline build '{build_id}'.")


@pipeline.group()
def snapshot() -> None:
    """Commands for pipeline snapshots."""


@snapshot.command("create", help="Create a snapshot of a pipeline.")
@click.argument("source")
@click.option(
    "--name",
    "-n",
    type=str,
    required=True,
    help="The name of the snapshot.",
)
@click.option(
    "--description",
    "-d",
    type=str,
    required=False,
    help="The description of the snapshot.",
)
@click.option(
    "--replace",
    "-r",
    is_flag=True,
    required=False,
    help="Whether to replace the existing snapshot with the same name.",
)
@click.option(
    "--tags",
    "-t",
    type=str,
    required=False,
    multiple=True,
    help="The tags to add to the snapshot.",
)
@click.option(
    "--config",
    "-c",
    "config_path",
    type=click.Path(exists=True, dir_okay=False),
    required=False,
    help="Path to configuration file for the snapshot.",
)
@click.option(
    "--stack",
    "-s",
    "stack_name_or_id",
    type=str,
    required=False,
    help="Name or ID of the stack to use for the snapshot.",
)
def create_pipeline_snapshot(
    source: str,
    name: str,
    description: Optional[str] = None,
    replace: Optional[bool] = None,
    tags: Optional[List[str]] = None,
    config_path: Optional[str] = None,
    stack_name_or_id: Optional[str] = None,
) -> None:
    """Create a snapshot of a pipeline.

    Args:
        source: Importable source resolving to a pipeline instance.
        name: Name of the snapshot.
        description: Description of the snapshot.
        tags: Tags to add to the snapshot.
        replace: Whether to replace the existing snapshot with the same name.
        config_path: Path to configuration file for the snapshot.
        stack_name_or_id: Name or ID of the stack for which the snapshot
            should be created.
    """
    if not Client().root:
        cli_utils.warning(
            "You're running the `zenml pipeline snapshot create` command "
            "without a ZenML repository. Your current working directory will "
            "be used as the source root relative to which the registered step "
            "classes will be resolved. To silence this warning, run `zenml "
            "init` at your source code root."
        )

    with cli_utils.temporary_active_stack(stack_name_or_id=stack_name_or_id):
        pipeline_instance = _import_pipeline(source=source)

        pipeline_instance = pipeline_instance.with_options(
            config_path=config_path
        )
        snapshot = pipeline_instance.create_snapshot(
            name=name,
            description=description,
            replace=replace,
            tags=tags,
        )

    cli_utils.declare(f"Created pipeline snapshot `{snapshot.name}`.")

    options = []

    if snapshot.runnable:
        options.append(
            "* run this snapshot from the dashboard or by calling "
            f"`zenml pipeline snapshot run {snapshot.id}`"
        )
    if snapshot.deployable:
        options.append(
            "* deploy this snapshot by calling "
            f"`zenml pipeline snapshot deploy {snapshot.id}`"
        )

    if options:
        cli_utils.declare("You can now:")
        cli_utils.declare("\n".join(options))


@snapshot.command("run", help="Run a snapshot.")
@click.argument("snapshot_name_or_id")
@click.option(
    "--pipeline",
    "-p",
    "pipeline_name_or_id",
    type=str,
    required=False,
    help="The name or ID of the pipeline.",
)
@click.option(
    "--config",
    "-c",
    "config_path",
    type=click.Path(exists=True, dir_okay=False),
    required=False,
    help="Path to configuration file for the run.",
)
def run_snapshot(
    snapshot_name_or_id: str,
    pipeline_name_or_id: Optional[str] = None,
    config_path: Optional[str] = None,
) -> None:
    """Run a snapshot.

    Args:
        snapshot_name_or_id: The name or ID of the snapshot to run.
        pipeline_name_or_id: The name or ID of the pipeline.
        config_path: Path to configuration file for the run.
    """
    snapshot = fetch_snapshot(snapshot_name_or_id, pipeline_name_or_id)
    try:
        run = Client().trigger_pipeline(
            snapshot_name_or_id=snapshot.id,
            config_path=config_path,
        )
        cli_utils.declare(f"Started snapshot run `{run.id}`.")
    except Exception as e:
        cli_utils.error(f"Failed to run snapshot: {e}")


@snapshot.command("deploy", help="Deploy a snapshot.")
@click.argument("snapshot_name_or_id")
@click.option(
    "--pipeline",
    "-p",
    "pipeline_name_or_id",
    type=str,
    required=False,
    help="The name or ID of the pipeline.",
)
@click.option(
    "--deployment",
    "-d",
    "deployment_name_or_id",
    type=str,
    required=False,
    help="The name or ID of the deployment to use for the pipeline. If "
    "not provided, the name of the snapshot or pipeline will be used. If an "
    "existing deployment with the same name already exists, an error will be "
    "raised, unless the --update or --overtake flag is used.",
)
@click.option(
    "--update",
    "-u",
    "update",
    is_flag=True,
    default=False,
    required=False,
    help="Update the deployment with the same name if it already exists.",
)
@click.option(
    "--overtake",
    "-o",
    "overtake",
    is_flag=True,
    default=False,
    required=False,
    help="Update the deployment with the same name if it already "
    "exists, even if it is owned by a different user.",
)
@click.option(
    "--timeout",
    "-t",
    "timeout",
    type=int,
    required=False,
    default=None,
    help="Maximum time in seconds to wait for the snapshot to be deployed.",
)
def deploy_snapshot(
    snapshot_name_or_id: str,
    pipeline_name_or_id: Optional[str] = None,
    deployment_name_or_id: Optional[str] = None,
    update: bool = False,
    overtake: bool = False,
    timeout: Optional[int] = None,
) -> None:
    """Deploy a pipeline for online inference.

    Args:
        snapshot_name_or_id: The name or ID of the snapshot to deploy.
        pipeline_name_or_id: The name or ID of the pipeline.
        deployment_name_or_id: Name or ID of the deployment to use for the
            pipeline.
        update: If True, update the deployment with the same name if it
            already exists.
        overtake: If True, update the deployment with the same name if
            it already exists, even if it is owned by a different user.
        timeout: The maximum time in seconds to wait for the pipeline to be
            deployed.
    """
    snapshot = fetch_snapshot(snapshot_name_or_id, pipeline_name_or_id)

    if not deployment_name_or_id:
        deployment_name_or_id = snapshot.name or snapshot.pipeline.name

    if not deployment_name_or_id:
        cli_utils.error(
            "No deployment name or ID provided. Please provide a deployment name or ID."
        )

    client = Client()
    try:
        deployment = client.get_deployment(deployment_name_or_id)
    except KeyError:
        pass
    else:
        if (
            deployment.user
            and deployment.user.id != client.active_user.id
            and not overtake
        ):
            confirmation = cli_utils.confirmation(
                f"Deployment with name or ID '{deployment_name_or_id}' is "
                f"owned by a different user '{deployment.user.name}'.\nDo you "
                "want to continue and provision it "
                "(hint: use the --overtake flag to skip this check)?"
            )
            if not confirmation:
                cli_utils.declare("Deployment provisioning canceled.")
                return

        elif (
            not update
            and not overtake
            and not uuid_utils.is_valid_uuid(deployment_name_or_id)
        ):
            confirmation = cli_utils.confirmation(
                f"Deployment with name or ID '{deployment_name_or_id}' already "
                "exists.\n"
                "Do you want to continue and update the existing "
                "deployment "
                "(hint: use the --update flag to skip this check) ?"
            )
            if not confirmation:
                cli_utils.declare("Deployment canceled.")
                return

    with console.status(
        f"Provisioning deployment '{deployment_name_or_id}'...\n"
    ):
        try:
            deployment = Client().provision_deployment(
                name_id_or_prefix=deployment_name_or_id,
                snapshot_id=snapshot.id,
                timeout=timeout,
            )
        except KeyError as e:
            cli_utils.exception(e)
        else:
            cli_utils.declare(
                f"Provisioned deployment '{deployment_name_or_id}'."
            )
            cli_utils.pretty_print_deployment(deployment, show_secret=True)

            dashboard_url = get_deployment_url(deployment)
            if dashboard_url:
                cli_utils.declare(
                    f"\nView in the ZenML UI: [link]{dashboard_url}[/link]",
                    no_wrap=True,
                )


@snapshot.command("list", help="List pipeline snapshots.")
@list_options(
    PipelineSnapshotFilter,
    default_columns=[
        "id",
        "name",
        "pipeline",
        "is_dynamic",
        "runnable",
        "deployable",
    ],
)
def list_pipeline_snapshots(
    columns: str, output_format: OutputFormat, **kwargs: Any
) -> None:
    """List all pipeline snapshots for the filter.

    Args:
        columns: Columns to display in output.
        output_format: Format for output (table/json/yaml/csv/tsv).
        **kwargs: Keyword arguments to filter pipeline snapshots.
    """
    # The helper returns default values for all filter fields even if not
    # explicitly provided by the user. But we only want to show named snapshots
    # by default in the CLI, so we need to explicitly set it to True in this
    # case.
    if kwargs.get("named_only") is None:
        kwargs["named_only"] = True

    client = Client()
    try:
        with console.status("Listing pipeline snapshots...\n"):
            pipeline_snapshots = client.list_snapshots(**kwargs)
    except KeyError as err:
        cli_utils.exception(err)
        return

    cli_utils.print_page(
        pipeline_snapshots,
        columns,
        output_format,
        empty_message="No pipeline snapshots found for this filter.",
    )


@snapshot.command("delete", help="Delete a pipeline snapshot.")
@click.argument("snapshot_name_or_id", type=str, required=True)
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    help="Don't ask for confirmation.",
)
def delete_pipeline_snapshot(
    snapshot_name_or_id: str, yes: bool = False
) -> None:
    """Delete a pipeline snapshot.

    Args:
        snapshot_name_or_id: The name or ID of the snapshot to delete.
        yes: If set, don't ask for confirmation.
    """
    if not yes:
        confirmation = cli_utils.confirmation(
            f"Are you sure you want to delete snapshot `{snapshot_name_or_id}`?"
        )
        if not confirmation:
            cli_utils.declare("Snapshot deletion canceled.")
            return

    try:
        Client().delete_snapshot(name_id_or_prefix=snapshot_name_or_id)
    except KeyError as e:
        cli_utils.exception(e)
    else:
        cli_utils.declare(f"Deleted snapshot '{snapshot_name_or_id}'.")



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Dynamic pipeline run context."""

import contextvars
import threading
from typing import TYPE_CHECKING

from typing_extensions import Self

from zenml.utils import context_utils

if TYPE_CHECKING:
    from zenml.execution.pipeline.dynamic.runner import DynamicPipelineRunner
    from zenml.models import PipelineRunResponse, PipelineSnapshotResponse
    from zenml.pipelines.dynamic.pipeline_definition import DynamicPipeline


class DynamicPipelineRunContext(context_utils.BaseContext):
    """Dynamic pipeline run context."""

    __context_var__ = contextvars.ContextVar("dynamic_pipeline_run_context")

    def __init__(
        self,
        pipeline: "DynamicPipeline",
        snapshot: "PipelineSnapshotResponse",
        run: "PipelineRunResponse",
        runner: "DynamicPipelineRunner",
    ) -> None:
        """Initialize the dynamic pipeline run context.

        Args:
            pipeline: The dynamic pipeline that is being executed.
            snapshot: The snapshot of the pipeline.
            run: The pipeline run.
            runner: The dynamic pipeline runner.
        """
        super().__init__()
        self._pipeline = pipeline
        self._snapshot = snapshot
        self._run = run
        self._runner = runner
        self._wait_condition_counter = 0
        self._wait_condition_counter_lock = threading.Lock()

    @property
    def pipeline(self) -> "DynamicPipeline":
        """The pipeline that is being executed.

        Returns:
            The pipeline that is being executed.
        """
        return self._pipeline

    @property
    def run(self) -> "PipelineRunResponse":
        """The pipeline run.

        Returns:
            The pipeline run.
        """
        return self._run

    @property
    def snapshot(self) -> "PipelineSnapshotResponse":
        """The snapshot of the pipeline.

        Returns:
            The snapshot of the pipeline.
        """
        return self._snapshot

    @property
    def runner(self) -> "DynamicPipelineRunner":
        """The runner executing the pipeline.

        Returns:
            The runner executing the pipeline.
        """
        return self._runner

    def next_wait_condition_name(self) -> str:
        """Get the next deterministic wait condition name for this run.

        Returns:
            A deterministic wait condition name.
        """
        with self._wait_condition_counter_lock:
            counter = self._wait_condition_counter
            self._wait_condition_counter += 1
        return f"wait_condition:{counter}"

    def __enter__(self) -> Self:
        """Enter the dynamic pipeline run context.

        Raises:
            RuntimeError: If the dynamic pipeline run context has already been
                entered.

        Returns:
            The dynamic pipeline run context object.
        """
        if self._token is not None:
            raise RuntimeError(
                "Calling a pipeline within a dynamic pipeline is not allowed."
            )
        return super().__enter__()



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the Evidently data validator."""

import os
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Optional,
    Sequence,
    Tuple,
    Type,
)

import pandas as pd

if TYPE_CHECKING:
    from evidently.legacy.pipeline.column_mapping import (  # type: ignore[import-untyped]
        ColumnMapping,
    )
    from evidently.legacy.report import Report  # type: ignore[import-untyped]
    from evidently.legacy.test_suite import (  # type: ignore[import-untyped]
        TestSuite,
    )

from zenml.data_validators import BaseDataValidator, BaseDataValidatorFlavor
from zenml.integrations.evidently.flavors.evidently_data_validator_flavor import (
    EvidentlyDataValidatorFlavor,
)
from zenml.logger import get_logger
from zenml.utils import source_utils

if TYPE_CHECKING:
    from zenml.integrations.evidently.metrics import EvidentlyMetricConfig
    from zenml.integrations.evidently.tests import EvidentlyTestConfig

logger = get_logger(__name__)


class EvidentlyDataValidator(BaseDataValidator):
    """Evidently data validator stack component."""

    NAME: ClassVar[str] = "Evidently"
    FLAVOR: ClassVar[Type[BaseDataValidatorFlavor]] = (
        EvidentlyDataValidatorFlavor
    )

    @classmethod
    def _unpack_options(
        cls, option_list: Sequence[Tuple[str, Dict[str, Any]]]
    ) -> Sequence[Any]:
        """Unpack Evidently options.

        Implements de-serialization for [Evidently options](https://docs.evidentlyai.com/metrics/customize_metric)
        that can be passed as constructor arguments when creating Profile and
        Dashboard objects. The convention used is that each item in the list
        consists of two elements:

        * a string containing the full class path of a `dataclass` based
        class with Evidently options
        * a dictionary with kwargs used as parameters for the option instance

        For example,

        ```python
            options = [
                (
                    "evidently.legacy.options.ColorOptions",{
                        "primary_color": "#5a86ad",
                        "fill_color": "#fff4f2",
                        "zero_line_color": "#016795",
                        "current_data_color": "#c292a1",
                        "reference_data_color": "#017b92",
                    }
                ),
            ]
        ```

        This is the same as saying:

        ```python
        from evidently.legacy.options import ColorOptions

        color_scheme = ColorOptions()
        color_scheme.primary_color = "#5a86ad"
        color_scheme.fill_color = "#fff4f2"
        color_scheme.zero_line_color = "#016795"
        color_scheme.current_data_color = "#c292a1"
        color_scheme.reference_data_color = "#017b92"
        ```

        Args:
            option_list: list of packed Evidently options

        Returns:
            A list of unpacked Evidently options

        Raises:
            ValueError: if one of the passed Evidently class paths cannot be
                resolved to an actual class.
        """
        options = []
        for option_clspath, option_args in option_list:
            try:
                option_cls = source_utils.load(option_clspath)
            except AttributeError:
                raise ValueError(
                    f"Could not map the `{option_clspath}` Evidently option "
                    f"class path to a valid class."
                )
            option = option_cls(**option_args)
            options.append(option)

        return options

    @staticmethod
    def _set_nltk_data_path() -> None:
        """Set the NLTK data path to the current working directory.

        This is necessary because the default download directory is not writable
        in some Docker containers.

        Raises:
            ImportError: if NLTK is not installed.
        """
        try:
            from nltk.data import (  # type: ignore[import-untyped]
                path as nltk_path,
            )
        except ImportError:
            raise ImportError(
                "NLTK is not installed. Please install NLTK to use "
                "Evidently text metrics and tests."
            )

        # Configure NLTK to use the current working directory to download and
        # lookup data. This is necessary because the default download directory
        # is not writable in some Docker containers.
        os.environ["NLTK_DATA"] = os.getcwd()
        nltk_path.append(os.getcwd())

    @staticmethod
    def _download_nltk_data() -> None:
        """Download NLTK data for text metrics and tests.

        Raises:
            ImportError: if NLTK is not installed.
        """
        try:
            import nltk  # type: ignore[import-untyped]
        except ImportError:
            raise ImportError(
                "NLTK is not installed. Please install NLTK to use "
                "Evidently text metrics and tests."
            )
        # Download NLTK data. We need this later on for the Evidently text report.
        nltk.download("words", download_dir=os.getcwd())
        nltk.download("wordnet", download_dir=os.getcwd())
        nltk.download("omw-1.4", download_dir=os.getcwd())

    def data_profiling(
        self,
        dataset: pd.DataFrame,
        comparison_dataset: Optional[pd.DataFrame] = None,
        profile_list: Optional[Sequence["EvidentlyMetricConfig"]] = None,
        column_mapping: Optional["ColumnMapping"] = None,
        report_options: Sequence[Tuple[str, Dict[str, Any]]] = [],
        download_nltk_data: bool = False,
        **kwargs: Any,
    ) -> "Report":
        """Analyze a dataset and generate a data report with Evidently.

        The method takes in an optional list of Evidently options to be passed
        to the report constructor (`report_options`). Each element in the list must be
        composed of two items: the first is a full class path of an Evidently
        option `dataclass`, the second is a dictionary of kwargs with the actual
        option parameters, e.g.:

        ```python
        options = [
            (
                "evidently.legacy.options.ColorOptions",{
                    "primary_color": "#5a86ad",
                    "fill_color": "#fff4f2",
                    "zero_line_color": "#016795",
                    "current_data_color": "#c292a1",
                    "reference_data_color": "#017b92",
                }
            ),
        ]
        ```

        Args:
            dataset: Target dataset to be profiled. When a comparison dataset
                is provided, this dataset is considered the reference dataset.
            comparison_dataset: Optional dataset to be used for data profiles
                that require a current dataset for comparison (e.g data drift
                profiles).
            profile_list: List of Evidently metric configurations to
                be included in the report. If not provided, all available
                metric presets will be included.
            column_mapping: Properties of the DataFrame columns used
            report_options: List of Evidently options to be passed to the
                report constructor.
            download_nltk_data: Whether to download NLTK data for text metrics.
                Defaults to False.
            **kwargs: Extra keyword arguments (unused).

        Returns:
            The Evidently Report as JSON object and as HTML.
        """
        from evidently.legacy.report import Report

        from zenml.integrations.evidently.metrics import EvidentlyMetricConfig

        self._set_nltk_data_path()
        if download_nltk_data:
            self._download_nltk_data()

        profile_list = profile_list or EvidentlyMetricConfig.default_metrics()
        metrics = [metric.to_evidently_metric() for metric in profile_list]

        unpacked_report_options = self._unpack_options(report_options)

        report = Report(metrics=metrics, options=unpacked_report_options)

        report.run(
            reference_data=dataset,
            current_data=comparison_dataset,
            column_mapping=column_mapping,
        )

        return report

    def data_validation(
        self,
        dataset: Any,
        comparison_dataset: Optional[Any] = None,
        check_list: Optional[Sequence["EvidentlyTestConfig"]] = None,
        test_options: Sequence[Tuple[str, Dict[str, Any]]] = [],
        column_mapping: Optional["ColumnMapping"] = None,
        download_nltk_data: bool = False,
        **kwargs: Any,
    ) -> "TestSuite":
        """Validate a dataset with Evidently.

        Args:
            dataset: Target dataset to be validated.
            comparison_dataset: Optional dataset to be used for data validation
                that require a baseline for comparison (e.g data drift
                validation).
            check_list: List of Evidently test configurations to be
                included in the test suite. If not provided, all available
                test presets will be included.
            test_options: List of Evidently options to be passed to the
                test suite constructor.
            column_mapping: Properties of the DataFrame columns used
            download_nltk_data: Whether to download NLTK data for text tests.
                Defaults to False.
            **kwargs: Extra keyword arguments (unused).

        Returns:
            The Evidently Test Suite as JSON object and as HTML.
        """
        from evidently.legacy.test_suite import TestSuite

        from zenml.integrations.evidently.tests import EvidentlyTestConfig

        if download_nltk_data:
            self._download_nltk_data()

        check_list = check_list or EvidentlyTestConfig.default_tests()
        tests = [test.to_evidently_test() for test in check_list]

        unpacked_test_options = self._unpack_options(test_options)

        test_suite = TestSuite(tests=tests, options=unpacked_test_options)
        test_suite.run(
            reference_data=dataset,
            current_data=comparison_dataset,
            column_mapping=column_mapping,
        )

        return test_suite



# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
#
# Parts of the `prepare_or_run_pipeline()` method of this file are
# inspired by the Kubernetes dag runner implementation of tfx
"""Kubernetes-native orchestrator."""

import json
import os
import random
import socket
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)
from uuid import UUID

from kubernetes import client as k8s_client
from kubernetes import config as k8s_config
from kubernetes.client import ApiException

from zenml.config.base_settings import BaseSettings
from zenml.constants import (
    METADATA_ORCHESTRATOR_RUN_ID,
    ORCHESTRATOR_DOCKER_IMAGE_KEY,
)
from zenml.enums import ExecutionMode, ExecutionStatus, StackComponentType
from zenml.integrations.kubernetes import kube_utils
from zenml.integrations.kubernetes.constants import (
    ENV_ZENML_KUBERNETES_RUN_ID,
    KUBERNETES_CRON_JOB_METADATA_KEY,
    KUBERNETES_SECRET_TOKEN_KEY_NAME,
    ORCHESTRATOR_ANNOTATION_KEY,
    STEP_NAME_ANNOTATION_KEY,
)
from zenml.integrations.kubernetes.flavors.kubernetes_orchestrator_flavor import (
    KubernetesOrchestratorConfig,
    KubernetesOrchestratorSettings,
)
from zenml.integrations.kubernetes.manifest_utils import (
    build_cron_job_manifest,
    build_job_manifest,
    build_pod_manifest,
    job_template_manifest_from_job,
    pod_template_manifest_from_pod,
)
from zenml.integrations.kubernetes.orchestrators.kubernetes_orchestrator_entrypoint_configuration import (
    KubernetesOrchestratorEntrypointConfiguration,
)
from zenml.integrations.kubernetes.pod_settings import KubernetesPodSettings
from zenml.logger import get_logger
from zenml.metadata.metadata_types import MetadataType
from zenml.models.v2.core.schedule import ScheduleUpdate
from zenml.orchestrators import (
    ContainerizedOrchestrator,
    PipelineSubmissionError,
    SubmissionResult,
)
from zenml.stack import StackValidator

if TYPE_CHECKING:
    from zenml.config.step_run_info import StepRunInfo
    from zenml.models import (
        PipelineRunResponse,
        PipelineSnapshotBase,
        PipelineSnapshotResponse,
        ScheduleResponse,
        StepRunResponse,
    )
    from zenml.stack import Stack

logger = get_logger(__name__)


class KubernetesOrchestrator(ContainerizedOrchestrator):
    """Orchestrator for running ZenML pipelines using native Kubernetes."""

    _k8s_client: Optional[k8s_client.ApiClient] = None

    def should_build_pipeline_image(
        self, snapshot: "PipelineSnapshotBase"
    ) -> bool:
        """Whether to always build the pipeline image.

        Args:
            snapshot: The pipeline snapshot.

        Returns:
            Whether to always build the pipeline image.
        """
        settings = cast(
            KubernetesOrchestratorSettings, self.get_settings(snapshot)
        )
        if settings.always_build_pipeline_image:
            return True
        else:
            return super().should_build_pipeline_image(snapshot)

    def get_kube_client(
        self, incluster: Optional[bool] = None
    ) -> k8s_client.ApiClient:
        """Getter for the Kubernetes API client.

        Args:
            incluster: Whether to use the in-cluster config or not. Overrides
                the `incluster` setting in the config.

        Returns:
            The Kubernetes API client.

        Raises:
            RuntimeError: if the Kubernetes connector behaves unexpectedly.
        """
        if incluster is None:
            incluster = self.config.incluster

        if incluster:
            try:
                kube_utils.load_kube_config(
                    incluster=incluster,
                )
                self._k8s_client = k8s_client.ApiClient()
                return self._k8s_client
            except Exception as e:
                if self.connector:
                    message = (
                        "Falling back to using the linked service connector "
                        "configuration."
                    )
                elif self.config.kubernetes_context:
                    message = (
                        f"Falling back to using the configured "
                        f"'{self.config.kubernetes_context}' kubernetes context."
                    )
                else:
                    raise RuntimeError(
                        f"The orchestrator failed to load the in-cluster "
                        f"Kubernetes configuration and there is no service "
                        f"connector or kubernetes_context to fall back to: {e}"
                    ) from e

                logger.debug(
                    f"Could not load the in-cluster Kubernetes configuration: "
                    f"{e}. {message}"
                )

        # Refresh the client also if the connector has expired
        if self._k8s_client and not self.connector_has_expired():
            return self._k8s_client

        connector = self.get_connector()
        if connector:
            client = connector.connect()
            if not isinstance(client, k8s_client.ApiClient):
                raise RuntimeError(
                    f"Expected a k8s_client.ApiClient while trying to use the "
                    f"linked connector, but got {type(client)}."
                )
            self._k8s_client = client
        else:
            kube_utils.load_kube_config(
                context=self.config.kubernetes_context,
            )
            self._k8s_client = k8s_client.ApiClient()

        return self._k8s_client

    @property
    def _k8s_core_api(self) -> k8s_client.CoreV1Api:
        """Getter for the Kubernetes Core API client.

        Returns:
            The Kubernetes Core API client.
        """
        return k8s_client.CoreV1Api(self.get_kube_client())

    @property
    def _k8s_batch_api(self) -> k8s_client.BatchV1Api:
        """Getter for the Kubernetes Batch API client.

        Returns:
            The Kubernetes Batch API client.
        """
        return k8s_client.BatchV1Api(self.get_kube_client())

    @property
    def _k8s_rbac_api(self) -> k8s_client.RbacAuthorizationV1Api:
        """Getter for the Kubernetes RBAC API client.

        Returns:
            The Kubernetes RBAC API client.
        """
        return k8s_client.RbacAuthorizationV1Api(self.get_kube_client())

    @property
    def config(self) -> KubernetesOrchestratorConfig:
        """Returns the `KubernetesOrchestratorConfig` config.

        Returns:
            The configuration.
        """
        return cast(KubernetesOrchestratorConfig, self._config)

    @property
    def settings_class(self) -> Optional[Type["BaseSettings"]]:
        """Settings class for the Kubernetes orchestrator.

        Returns:
            The settings class.
        """
        return KubernetesOrchestratorSettings

    def get_kubernetes_contexts(self) -> Tuple[List[str], str]:
        """Get list of configured Kubernetes contexts and the active context.

        Raises:
            RuntimeError: if the Kubernetes configuration cannot be loaded.

        Returns:
            context_name: List of configured Kubernetes contexts
            active_context_name: Name of the active Kubernetes context.
        """
        try:
            contexts, active_context = k8s_config.list_kube_config_contexts()
        except k8s_config.config_exception.ConfigException as e:
            raise RuntimeError(
                "Could not load the Kubernetes configuration"
            ) from e

        context_names = [c["name"] for c in contexts]
        active_context_name = active_context["name"]
        return context_names, active_context_name

    @property
    def validator(self) -> Optional[StackValidator]:
        """Defines the validator that checks whether the stack is valid.

        Returns:
            Stack validator.
        """

        def _validate_local_requirements(stack: "Stack") -> Tuple[bool, str]:
            """Validates that the stack contains no local components.

            Args:
                stack: The stack.

            Returns:
                Whether the stack is valid or not.
                An explanation why the stack is invalid, if applicable.
            """
            container_registry = stack.container_registry

            # should not happen, because the stack validation takes care of
            # this, but just in case
            assert container_registry is not None

            kubernetes_context = self.config.kubernetes_context
            msg = f"'{self.name}' Kubernetes orchestrator error: "

            if not self.connector:
                if self.config.incluster:
                    # No service connector or kubernetes_context is needed when
                    # the orchestrator is being used from within a Kubernetes
                    # cluster.
                    pass
                elif kubernetes_context:
                    contexts, active_context = self.get_kubernetes_contexts()

                    if kubernetes_context not in contexts:
                        return False, (
                            f"{msg}could not find a Kubernetes context named "
                            f"'{kubernetes_context}' in the local "
                            "Kubernetes configuration. Please make sure that "
                            "the Kubernetes cluster is running and that the "
                            "kubeconfig file is configured correctly. To list "
                            "all configured contexts, run:\n\n"
                            "  `kubectl config get-contexts`\n"
                        )
                    if kubernetes_context != active_context:
                        logger.warning(
                            f"{msg}the Kubernetes context "  # nosec
                            f"'{kubernetes_context}' configured for the "
                            f"Kubernetes orchestrator is not the same as the "
                            f"active context in the local Kubernetes "
                            f"configuration. If this is not deliberate, you "
                            f"should update the orchestrator's "
                            f"`kubernetes_context` field by running:\n\n"
                            f"  `zenml orchestrator update {self.name} "
                            f"--kubernetes_context={active_context}`\n"
                            f"To list all configured contexts, run:\n\n"
                            f"  `kubectl config get-contexts`\n"
                            f"To set the active context to be the same as the "
                            f"one configured in the Kubernetes orchestrator "
                            f"and silence this warning, run:\n\n"
                            f"  `kubectl config use-context "
                            f"{kubernetes_context}`\n"
                        )
                else:
                    return False, (
                        f"{msg}you must either link this orchestrator to a "
                        "Kubernetes service connector (see the 'zenml "
                        "orchestrator connect' CLI command), explicitly set "
                        "the `kubernetes_context` attribute to the name of the "
                        "Kubernetes config context pointing to the cluster "
                        "where you would like to run pipelines, or set the "
                        "`incluster` attribute to `True`."
                    )

            silence_local_validations_msg = (
                f"To silence this warning, set the "
                f"`skip_local_validations` attribute to True in the "
                f"orchestrator configuration by running:\n\n"
                f"  'zenml orchestrator update {self.name} "
                f"--skip_local_validations=True'\n"
            )

            if (
                not self.config.skip_local_validations
                and not self.config.is_local
            ):
                # if the orchestrator is not running in a local k3d cluster,
                # we cannot have any other local components in our stack,
                # because we cannot mount the local path into the container.
                # This may result in problems when running the pipeline, because
                # the local components will not be available inside the
                # kubernetes containers.

                # go through all stack components and identify those that
                # advertise a local path where they persist information that
                # they need to be available when running pipelines.
                for stack_comp in stack.components.values():
                    if stack_comp.local_path is None:
                        continue
                    return False, (
                        f"{msg}the Kubernetes orchestrator is configured to "
                        f"run pipelines in a remote Kubernetes cluster but the "
                        f"'{stack_comp.name}' {stack_comp.type.value} "
                        f"is a local stack component "
                        f"and will not be available in the Kubernetes pipeline "
                        f"step.\nPlease ensure that you always use non-local "
                        f"stack components with a remote Kubernetes orchestrator, "
                        f"otherwise you may run into pipeline execution "
                        f"problems. You should use a flavor of "
                        f"{stack_comp.type.value} other than "
                        f"'{stack_comp.flavor}'.\n"
                        + silence_local_validations_msg
                    )

                # if the orchestrator is remote, the container registry must
                # also be remote.
                if container_registry.config.is_local:
                    return False, (
                        f"{msg}the Kubernetes orchestrator is configured to "
                        "run pipelines in a remote Kubernetes cluster but the "
                        f"'{container_registry.name}' container registry URI "
                        f"'{container_registry.config.uri}' "
                        f"points to a local container registry. Please ensure "
                        f"that you always use non-local stack components with "
                        f"a remote Kubernetes orchestrator, otherwise you will "
                        f"run into problems. You should use a flavor of "
                        f"container registry other than "
                        f"'{container_registry.flavor}'.\n"
                        + silence_local_validations_msg
                    )

            return True, ""

        return StackValidator(
            required_components={
                StackComponentType.CONTAINER_REGISTRY,
                StackComponentType.IMAGE_BUILDER,
            },
            custom_validation_function=_validate_local_requirements,
        )

    def get_token_secret_name(self, snapshot_id: UUID) -> str:
        """Returns the name of the secret that contains the ZenML token.

        Args:
            snapshot_id: The ID of the snapshot.

        Returns:
            The name of the secret that contains the ZenML token.
        """
        return f"zenml-token-{snapshot_id}"

    @property
    def supported_execution_modes(self) -> List[ExecutionMode]:
        """Returns the supported execution modes for this flavor.

        Returns:
            A tuple of supported execution modes.
        """
        return [
            ExecutionMode.FAIL_FAST,
            ExecutionMode.STOP_ON_FAILURE,
            ExecutionMode.CONTINUE_ON_FAILURE,
        ]

    def submit_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        base_environment: Dict[str, str],
        step_environments: Dict[str, Dict[str, str]],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submit a static pipeline to the orchestrator.

        Args:
            snapshot: The pipeline snapshot to submit.
            stack: The stack the pipeline will run on.
            base_environment: Base environment shared by all steps. This should
                be set if your orchestrator for example runs one container that
                is responsible for starting all the steps.
            step_environments: Environment variables to set when executing
                specific steps.
            placeholder_run: An optional placeholder run for the snapshot.

        Returns:
            Optional submission result.
        """
        for step_name, step in snapshot.step_configurations.items():
            if self.requires_resources_in_orchestration_environment(step):
                logger.warning(
                    "Specifying step resources is not yet supported for "
                    "the Kubernetes orchestrator, ignoring resource "
                    "configuration for step %s.",
                    step_name,
                )

            if retry_config := step.config.retry:
                if retry_config.delay or retry_config.backoff:
                    logger.warning(
                        "Specifying retry delay or backoff is not supported "
                        "for the Kubernetes orchestrator."
                    )

        command = KubernetesOrchestratorEntrypointConfiguration.get_entrypoint_command()
        args = KubernetesOrchestratorEntrypointConfiguration.get_entrypoint_arguments(
            snapshot_id=snapshot.id,
            run_id=placeholder_run.id if placeholder_run else None,
        )

        result = self._submit_orchestrator_job(
            snapshot=snapshot,
            command=command,
            args=args,
            environment=base_environment,
            placeholder_run=placeholder_run,
        )

        return result

    def submit_dynamic_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        environment: Dict[str, str],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a dynamic pipeline to the orchestrator.

        Args:
            snapshot: The snapshot of the pipeline.
            stack: The stack to use for the pipeline.
            environment: The environment variables to set in the pipeline.
            placeholder_run: The placeholder run for the pipeline.

        Returns:
            Optional submission result.
        """
        from zenml.pipelines.dynamic.entrypoint_configuration import (
            DynamicPipelineEntrypointConfiguration,
        )

        command = (
            DynamicPipelineEntrypointConfiguration.get_entrypoint_command()
        )
        args = DynamicPipelineEntrypointConfiguration.get_entrypoint_arguments(
            snapshot_id=snapshot.id,
            run_id=placeholder_run.id if placeholder_run else None,
        )

        return self._submit_orchestrator_job(
            snapshot=snapshot,
            command=command,
            args=args,
            environment=environment,
            placeholder_run=placeholder_run,
        )

    def _prepare_job_manifest(
        self,
        name: str,
        command: List[str],
        args: List[str],
        image: str,
        environment: Dict[str, str],
        labels: Dict[str, str],
        annotations: Dict[str, str],
        settings: KubernetesOrchestratorSettings,
        pod_settings: Optional[KubernetesPodSettings] = None,
        backoff_limit: Optional[int] = None,
    ) -> k8s_client.V1Job:
        """Prepares the job manifest for a Kubernetes job.

        Args:
            name: The name of the job.
            command: The command to run in the job.
            args: The arguments to pass to the job.
            image: The image to use for the job.
            environment: The environment variables to set in the job.
            labels: The labels to add to the job.
            annotations: The annotations to add to the job.
            settings: Component settings for the orchestrator.
            pod_settings: Optional settings for the pod.
            backoff_limit: The backoff limit for the job.

        Returns:
            The job manifest.
        """
        # Authorize pod to run Kubernetes commands inside the cluster.
        service_account_name = self._get_service_account_name(settings)

        # We set some default minimum resource requests for the orchestrator pod
        # here if the user has not specified any, because the orchestrator pod
        # takes up some memory resources itself and, if not specified, the pod
        # will be scheduled on any node regardless of available memory and risk
        # negatively impacting or even crashing the node due to memory pressure.
        pod_settings = kube_utils.apply_default_resource_requests(
            memory="400Mi",
            cpu="100m",
            pod_settings=pod_settings,
        )

        pod_manifest = build_pod_manifest(
            pod_name=None,
            image_name=image,
            command=command,
            args=args,
            privileged=False,
            pod_settings=pod_settings,
            service_account_name=service_account_name,
            env=environment,
            labels=labels,
            mount_local_stores=self.config.is_local,
            termination_grace_period_seconds=settings.pod_stop_grace_period,
        )

        pod_failure_policy = settings.pod_failure_policy or {
            # These rules are applied sequentially. This means any failure in
            # the main container will count towards the max retries. Any other
            # disruption will not count towards the max retries.
            "rules": [
                # If the main container fails, we count it towards the max
                # retries.
                {
                    "action": "Count",
                    "onExitCodes": {
                        "containerName": "main",
                        "operator": "NotIn",
                        "values": [0],
                    },
                },
                # If the pod is interrupted at any other time, we don't count
                # it as a retry
                {
                    "action": "Ignore",
                    "onPodConditions": [
                        {
                            "type": "DisruptionTarget",
                            "status": "True",
                        }
                    ],
                },
            ]
        }

        return build_job_manifest(
            job_name=name,
            pod_template=pod_template_manifest_from_pod(pod_manifest),
            backoff_limit=backoff_limit,
            ttl_seconds_after_finished=settings.ttl_seconds_after_finished,
            active_deadline_seconds=settings.active_deadline_seconds,
            pod_failure_policy=pod_failure_policy,
            labels=labels,
            annotations=annotations,
        )

    def _get_job_name(
        self,
        settings: KubernetesOrchestratorSettings,
        pipeline_name: str,
        step_name: Optional[str] = None,
    ) -> str:
        """Gets a job name for a Kubernetes job.

        Args:
            settings: The settings for the orchestrator.
            pipeline_name: The name of the pipeline.
            step_name: The name of the step.

        Returns:
            The job name.
        """
        job_name = settings.job_name_prefix or ""
        random_prefix = "".join(random.choices("0123456789abcdef", k=8))
        job_name += f"-{random_prefix}-{pipeline_name}"
        if step_name:
            job_name += f"-{step_name}"
        # The job name will be used as a label on the pods, so we need to make
        # sure it doesn't exceed the label length limit
        job_name = kube_utils.sanitize_label(job_name)
        return job_name

    @contextmanager
    def _create_auth_secret_if_necessary(
        self,
        snapshot: "PipelineSnapshotResponse",
        environment: Dict[str, str],
        pod_settings: KubernetesPodSettings,
    ) -> Generator[None, None, None]:
        """Creates an authentication secret if necessary.

        If the authentication secret is created and some exception is raised,
        the secret will be deleted.

        Args:
            snapshot: The pipeline snapshot.
            environment: The environment variables to set.
            pod_settings: The pod settings to update.

        Raises:
            Exception: If an exception happens while the context manager is
                active.

        Yields:
            None.
        """
        try:
            if self.config.pass_zenml_token_as_secret:
                secret_name = self.get_token_secret_name(snapshot.id)
                token = environment.pop("ZENML_STORE_API_TOKEN")
                kube_utils.create_or_update_secret(
                    core_api=self._k8s_core_api,
                    namespace=self.config.kubernetes_namespace,
                    secret_name=secret_name,
                    data={KUBERNETES_SECRET_TOKEN_KEY_NAME: token},
                )
                pod_settings.env.append(
                    {
                        "name": "ZENML_STORE_API_TOKEN",
                        "valueFrom": {
                            "secretKeyRef": {
                                "name": secret_name,
                                "key": KUBERNETES_SECRET_TOKEN_KEY_NAME,
                            }
                        },
                    }
                )
            yield
        except Exception as e:
            if self.config.pass_zenml_token_as_secret:
                secret_name = self.get_token_secret_name(snapshot.id)
                try:
                    kube_utils.delete_secret(
                        core_api=self._k8s_core_api,
                        namespace=self.config.kubernetes_namespace,
                        secret_name=secret_name,
                    )
                except Exception as cleanup_error:
                    logger.error(
                        "Error cleaning up secret %s: %s",
                        secret_name,
                        cleanup_error,
                    )
            raise e

    def _submit_orchestrator_job(
        self,
        snapshot: "PipelineSnapshotResponse",
        command: List[str],
        args: List[str],
        environment: Dict[str, str],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits an orchestrator job to Kubernetes.

        Args:
            snapshot: The pipeline snapshot.
            command: The command to run in the job.
            args: The arguments to pass to the job.
            environment: The environment variables to set in the job.
            placeholder_run: The placeholder run for the job.

        Raises:
            RuntimeError: If a schedule without cron expression is given.
            PipelineSubmissionError: If an APIException happens during job submission,
                it is propagated as custom error.

        Returns:
            Optional submission result.
        """
        pipeline_name = snapshot.pipeline_configuration.name
        settings = cast(
            KubernetesOrchestratorSettings, self.get_settings(snapshot)
        )
        orchestrator_pod_settings = (
            settings.orchestrator_pod_settings or KubernetesPodSettings()
        )

        try:
            image = self.get_image(snapshot=snapshot)
        except KeyError:
            # If no generic pipeline image exists (which means all steps of a
            # static pipeline have custom builds) we use a random step image as
            # all of them include dependencies for the active stack
            invocation_id = next(iter(snapshot.step_configurations))
            image = self.get_image(snapshot=snapshot, step_name=invocation_id)

        labels = {
            "project_id": kube_utils.sanitize_label(str(snapshot.project_id)),
            "pipeline": kube_utils.sanitize_label(pipeline_name),
        }

        if placeholder_run:
            labels["run_id"] = kube_utils.sanitize_label(
                str(placeholder_run.id)
            )
            labels["run_name"] = kube_utils.sanitize_label(
                placeholder_run.name
            )

        annotations = {
            ORCHESTRATOR_ANNOTATION_KEY: str(self.id),
        }

        job_name = self._get_job_name(
            settings, pipeline_name=snapshot.pipeline_configuration.name
        )

        try:
            with self._create_auth_secret_if_necessary(
                snapshot, environment, orchestrator_pod_settings
            ):
                job_manifest = self._prepare_job_manifest(
                    name=job_name,
                    command=command,
                    args=args,
                    image=image,
                    environment=environment,
                    labels=labels,
                    annotations=annotations,
                    settings=settings,
                    pod_settings=orchestrator_pod_settings,
                    backoff_limit=settings.orchestrator_job_backoff_limit,
                )

                if snapshot.schedule:
                    if not snapshot.schedule.cron_expression:
                        raise RuntimeError(
                            "The Kubernetes orchestrator only supports scheduling via "
                            "CRON jobs, but the run was configured with a manual "
                            "schedule. Use `Schedule(cron_expression=...)` instead."
                        )
                    cron_expression = snapshot.schedule.cron_expression
                    cron_job_manifest = build_cron_job_manifest(
                        cron_expression=cron_expression,
                        job_template=job_template_manifest_from_job(
                            job_manifest
                        ),
                        successful_jobs_history_limit=settings.successful_jobs_history_limit,
                        failed_jobs_history_limit=settings.failed_jobs_history_limit,
                        concurrency_policy=settings.concurrency_policy,
                        starting_deadline_seconds=settings.starting_deadline_seconds,
                    )

                    cron_job = self._k8s_batch_api.create_namespaced_cron_job(
                        body=cron_job_manifest,
                        namespace=self.config.kubernetes_namespace,
                    )
                    logger.info(
                        f"Created Kubernetes CronJob `{cron_job.metadata.name}` "
                        f"with CRON expression `{cron_expression}`."
                    )
                    return SubmissionResult(
                        metadata={
                            KUBERNETES_CRON_JOB_METADATA_KEY: cron_job.metadata.name,
                        }
                    )
                else:
                    kube_utils.create_job(
                        batch_api=self._k8s_batch_api,
                        namespace=self.config.kubernetes_namespace,
                        job_manifest=job_manifest,
                        api_request_timeout=settings.api_request_timeout,
                    )

                    if settings.synchronous:

                        def _wait_for_run_to_finish() -> None:
                            logger.info(
                                "Waiting for orchestrator job to finish..."
                            )
                            kube_utils.wait_for_job_to_finish(
                                get_client=self.get_kube_client,
                                namespace=self.config.kubernetes_namespace,
                                job_name=job_name,
                                backoff_interval=settings.job_monitoring_interval,
                                fail_on_container_waiting_reasons=settings.fail_on_container_waiting_reasons,
                                stream_logs=True,
                                api_request_timeout=settings.api_request_timeout,
                            )

                        return SubmissionResult(
                            wait_for_completion=_wait_for_run_to_finish,
                        )
                    else:
                        logger.info(
                            f"Orchestrator job `{job_name}` started. "
                            f"Run the following command to inspect the logs: "
                            f"`kubectl -n {self.config.kubernetes_namespace} logs "
                            f"job/{job_name}`"
                        )
                        return None

        except ApiException as e:
            body = json.loads(e.body or "{}")
            raise PipelineSubmissionError(
                f"Orchestrator failed to submit Kubernetes job with '{e.reason}' ({e.status}). "
                f"{body.get('message', '')}"
            )

    def submit_isolated_step(
        self, step_run_info: "StepRunInfo", environment: Dict[str, str]
    ) -> None:
        """Submit an isolated step.

        Args:
            step_run_info: The step run information.
            environment: The environment variables to set in the execution
                environment.
        """
        from zenml.step_operators.step_operator_entrypoint_configuration import (
            StepOperatorEntrypointConfiguration,
        )

        logger.info(
            "Launching job for step `%s`.",
            step_run_info.pipeline_step_name,
        )

        settings = cast(
            KubernetesOrchestratorSettings, self.get_settings(step_run_info)
        )
        image = step_run_info.get_image(key=ORCHESTRATOR_DOCKER_IMAGE_KEY)
        command = StepOperatorEntrypointConfiguration.get_entrypoint_command()
        args = StepOperatorEntrypointConfiguration.get_entrypoint_arguments(
            step_name=step_run_info.pipeline_step_name,
            snapshot_id=(step_run_info.snapshot.id),
            step_run_id=str(step_run_info.step_run_id),
        )

        labels = {
            "project_id": kube_utils.sanitize_label(
                str(step_run_info.snapshot.project_id)
            ),
            "pipeline": kube_utils.sanitize_label(step_run_info.pipeline.name),
            "run_id": kube_utils.sanitize_label(str(step_run_info.run_id)),
            "run_name": kube_utils.sanitize_label(str(step_run_info.run_name)),
            "step_run_id": kube_utils.sanitize_label(
                str(step_run_info.step_run_id)
            ),
            "step_name": kube_utils.sanitize_label(
                step_run_info.pipeline_step_name
            ),
        }
        annotations = {
            STEP_NAME_ANNOTATION_KEY: step_run_info.pipeline_step_name,
        }

        job_name = self._get_job_name(
            settings,
            pipeline_name=step_run_info.pipeline.name,
            step_name=step_run_info.pipeline_step_name,
        )

        job_manifest = self._prepare_job_manifest(
            name=job_name,
            command=command,
            args=args,
            image=image,
            environment=environment,
            labels=labels,
            annotations=annotations,
            settings=settings,
            pod_settings=settings.pod_settings,
            # In the dynamic pipeline case, we can't handle retries at the
            # orchestrator level because the entrypoint args contain a step
            # run ID.
            backoff_limit=settings.backoff_limit_margin,
        )

        kube_utils.create_job(
            batch_api=self._k8s_batch_api,
            namespace=self.config.kubernetes_namespace,
            job_manifest=job_manifest,
            api_request_timeout=settings.api_request_timeout,
        )

    def get_isolated_step_status(
        self, step_run: "StepRunResponse"
    ) -> ExecutionStatus:
        """Get the status of an isolated step.

        Args:
            step_run: The step run.

        Returns:
            The status.
        """
        label_selector = (
            f"step_run_id={kube_utils.sanitize_label(str(step_run.id))}"
        )
        try:
            job_list = kube_utils.list_jobs(
                batch_api=self._k8s_batch_api,
                namespace=self.config.kubernetes_namespace,
                label_selector=label_selector,
                api_request_timeout=self.config.api_request_timeout,
            )
        except Exception as e:
            logger.warning(
                "Failed to list jobs for step run `%s`: %s", step_run.id, e
            )
            return ExecutionStatus.FAILED

        if not job_list.items:
            logger.warning("No jobs found for step run `%s`", step_run.id)
            return ExecutionStatus.FAILED

        job = job_list.items[0]
        if job.status.conditions:
            for condition in job.status.conditions:
                if condition.type == "Complete" and condition.status == "True":
                    return ExecutionStatus.COMPLETED
                if condition.type == "Failed" and condition.status == "True":
                    return ExecutionStatus.FAILED

        return ExecutionStatus.RUNNING

    def stop_isolated_step(self, step_run: "StepRunResponse") -> None:
        """Stop an isolated step.

        Args:
            step_run: The step run.
        """
        label_selector = (
            f"step_run_id={kube_utils.sanitize_label(str(step_run.id))}"
        )
        try:
            job_list = kube_utils.list_jobs(
                batch_api=self._k8s_batch_api,
                namespace=self.config.kubernetes_namespace,
                label_selector=label_selector,
                api_request_timeout=self.config.api_request_timeout,
            )
        except Exception as e:
            logger.warning(
                "Failed to list jobs for step run `%s`: %s", step_run.id, e
            )
            return

        if not job_list.items:
            logger.warning("No jobs found for step run `%s`", step_run.id)
            return

        job_name = job_list.items[0].metadata.name
        self._k8s_batch_api.delete_namespaced_job(
            name=job_name,
            namespace=self.config.kubernetes_namespace,
            propagation_policy="Foreground",
        )
        logger.info(f"Successfully stopped step job: {job_name}")

    def _get_service_account_name(
        self, settings: KubernetesOrchestratorSettings
    ) -> str:
        """Returns the service account name to use for the orchestrator pod.

        If the user has not specified a service account name in the settings,
        we create a new service account with the required permissions.

        Args:
            settings: The orchestrator settings.

        Returns:
            The service account name.
        """
        if settings.service_account_name:
            return settings.service_account_name
        else:
            service_account_name = "zenml-service-account"
            kube_utils.create_edit_service_account(
                core_api=self._k8s_core_api,
                rbac_api=self._k8s_rbac_api,
                service_account_name=service_account_name,
                namespace=self.config.kubernetes_namespace,
            )
            return service_account_name

    def get_orchestrator_run_id(self) -> str:
        """Returns the active orchestrator run id.

        Returns:
            The orchestrator run id.
        """
        try:
            return os.environ[ENV_ZENML_KUBERNETES_RUN_ID]
        except KeyError:
            # This means we're in a dynamic pipeline orchestration container.
            # In case the orchestrator pod is restarted, we need to return the
            # same `orchestrator_run_id` as the previous pod, so we use the
            # orchestrator job name (which created this pod and handles the
            # restarts) as the unique ID.
            pod_name = socket.gethostname()
            orchestrator_run_id = None
            try:
                orchestrator_run_id = kube_utils.get_parent_job_name(
                    core_api=self._k8s_core_api,
                    pod_name=pod_name,
                    namespace=self.config.kubernetes_namespace,
                    api_request_timeout=self.config.api_request_timeout,
                )
            except Exception as e:
                logger.warning(
                    "Failed to fetch orchestrator job name. Using pod name as "
                    "fallback, which may cause issues if the orchestrator pod "
                    "is restarted. Error: %s",
                    e,
                )
            return orchestrator_run_id or pod_name

    def _stop_run(
        self, run: "PipelineRunResponse", graceful: bool = True
    ) -> None:
        """Stops a specific pipeline run by terminating step pods.

        Args:
            run: The run that was executed by this orchestrator.
            graceful: If True, does nothing (lets the orchestrator and steps finish naturally).
                If False, stops all running step jobs.

        Raises:
            RuntimeError: If we fail to stop the run.
        """
        # If graceful, do nothing and let the orchestrator handle the stop naturally
        if graceful:
            logger.info(
                "Graceful stop requested - the orchestrator pod will handle "
                "stopping naturally"
            )
            return

        jobs_stopped = []
        errors = []

        # Find all jobs running steps of the pipeline
        label_selector = f"run_id={kube_utils.sanitize_label(str(run.id))}"
        try:
            job_list = kube_utils.list_jobs(
                batch_api=self._k8s_batch_api,
                namespace=self.config.kubernetes_namespace,
                label_selector=label_selector,
                api_request_timeout=self.config.api_request_timeout,
            )
        except Exception as e:
            raise RuntimeError(
                f"Failed to list step jobs with run ID {run.id}: {e}"
            )

        for job in job_list.items:
            if not kube_utils.is_step_job(job):
                # This is the orchestrator job which stops by itself
                continue

            if job.status and job.status.conditions:
                # Don't delete completed/failed jobs
                for condition in job.status.conditions:
                    if (
                        condition.type == "Complete"
                        and condition.status == "True"
                    ):
                        continue
                    if (
                        condition.type == "Failed"
                        and condition.status == "True"
                    ):
                        continue

            try:
                self._k8s_batch_api.delete_namespaced_job(
                    name=job.metadata.name,
                    namespace=self.config.kubernetes_namespace,
                    propagation_policy="Foreground",
                )
                jobs_stopped.append(f"step job: {job.metadata.name}")
                logger.debug(
                    f"Successfully initiated graceful stop of step job: {job.metadata.name}"
                )
            except Exception as e:
                error_msg = f"Failed to stop step job {job.metadata.name}: {e}"
                logger.warning(error_msg)
                errors.append(error_msg)

        # Summary logging
        settings = cast(KubernetesOrchestratorSettings, self.get_settings(run))
        grace_period_seconds = settings.pod_stop_grace_period
        if jobs_stopped:
            logger.debug(
                f"Successfully initiated graceful termination of: {', '.join(jobs_stopped)}. "
                f"Pods will terminate within {grace_period_seconds} seconds."
            )

        if errors:
            error_summary = "; ".join(errors)
            if not jobs_stopped:
                # If nothing was stopped successfully, raise an error
                raise RuntimeError(
                    f"Failed to stop pipeline run: {error_summary}"
                )
            else:
                # If some things were stopped but others failed, raise an error
                raise RuntimeError(
                    f"Partial stop operation completed with errors: {error_summary}"
                )

        if not jobs_stopped and not errors:
            logger.info(
                f"No running step jobs found for pipeline run with ID: {run.id}"
            )

    def fetch_status(
        self, run: "PipelineRunResponse", include_steps: bool = False
    ) -> Tuple[
        Optional[ExecutionStatus], Optional[Dict[str, ExecutionStatus]]
    ]:
        """Refreshes the status of a specific pipeline run.

        Args:
            run: The run that was executed by this orchestrator.
            include_steps: If True, also fetch the status of individual steps.

        Returns:
            A tuple of (pipeline_status, step_statuses).
            If include_steps is False, step_statuses will be None.
            If include_steps is True, step_statuses will be a dict (possibly empty).
        """
        pipeline_status = None
        include_run_status = not run.status.is_finished

        label_selector = f"run_id={kube_utils.sanitize_label(str(run.id))}"
        try:
            job_list = kube_utils.list_jobs(
                batch_api=self._k8s_batch_api,
                namespace=self.config.kubernetes_namespace,
                label_selector=label_selector,
                api_request_timeout=self.config.api_request_timeout,
            )
        except Exception as e:
            logger.warning(f"Failed to list jobs for run {run.id}: {e}")
            return None, None

        step_statuses = {}
        # Only fetch steps if we really need them
        steps_dict = run.steps if include_steps else {}

        for job in job_list.items:
            if not job.metadata or not job.metadata.annotations:
                continue

            is_orchestrator_job = (
                ORCHESTRATOR_ANNOTATION_KEY in job.metadata.annotations
            )
            if is_orchestrator_job:
                if include_run_status:
                    pipeline_status = self._map_job_status_to_execution_status(
                        job
                    )
                continue

            step_name = job.metadata.annotations.get(
                STEP_NAME_ANNOTATION_KEY, None
            )
            if not include_steps or not step_name:
                continue

            step_response = steps_dict.get(step_name, None)

            if step_response is None:
                continue

            # If the step is already in a finished state, skip
            if step_response and step_response.status.is_finished:
                continue

            execution_status = self._map_job_status_to_execution_status(job)
            if execution_status is not None:
                step_statuses[step_name] = execution_status

        return pipeline_status, step_statuses

    def _map_job_status_to_execution_status(
        self, job: k8s_client.V1Job
    ) -> Optional[ExecutionStatus]:
        """Map Kubernetes job status to ZenML execution status.

        Args:
            job: The Kubernetes job.

        Returns:
            The corresponding ZenML execution status, or None if no clear status.
        """
        if job.status and job.status.conditions:
            for condition in job.status.conditions:
                if condition.type == "Complete" and condition.status == "True":
                    return ExecutionStatus.COMPLETED
                elif condition.type == "Failed" and condition.status == "True":
                    return ExecutionStatus.FAILED

        # Return None if no clear status - don't update
        return None

    def get_pipeline_run_metadata(
        self, run_id: UUID
    ) -> Dict[str, "MetadataType"]:
        """Get general component-specific metadata for a pipeline run.

        Args:
            run_id: The ID of the pipeline run.

        Returns:
            A dictionary of metadata.
        """
        return {
            METADATA_ORCHESTRATOR_RUN_ID: self.get_orchestrator_run_id(),
        }

    def update_schedule(
        self, schedule: "ScheduleResponse", update: ScheduleUpdate
    ) -> None:
        """Updates a schedule.

        Args:
            schedule: The schedule to update.
            update: The update to apply to the schedule.

        Raises:
            RuntimeError: If the cron job name is not found.
        """
        cron_job_name = schedule.run_metadata.get(
            KUBERNETES_CRON_JOB_METADATA_KEY
        )
        if not cron_job_name:
            raise RuntimeError("Unable to find cron job name for schedule.")

        spec = {}

        if update.cron_expression:
            spec["schedule"] = update.cron_expression

        if update.active is not None:
            spec["suspend"] = not update.active  # type: ignore[assignment]

        if spec:
            self._k8s_batch_api.patch_namespaced_cron_job(
                name=cron_job_name,
                namespace=self.config.kubernetes_namespace,
                body={"spec": spec},
            )

    def delete_schedule(self, schedule: "ScheduleResponse") -> None:
        """Deletes a schedule.

        Args:
            schedule: The schedule to delete.

        Raises:
            ApiException: Kubernetes API error (other than 404) if deletion fails.
        """
        cron_job_name = schedule.run_metadata.get(
            KUBERNETES_CRON_JOB_METADATA_KEY
        )

        if not cron_job_name:
            logger.warning(
                "Unable to find cron job for schedule %s.",
                schedule.name,
            )
            return

        try:
            self._k8s_batch_api.delete_namespaced_cron_job(
                name=cron_job_name,
                namespace=self.config.kubernetes_namespace,
            )
        except ApiException as e:
            if e.status == 404:
                logger.warning(
                    "Unable to find cron job %s for schedule %s.",
                    cron_job_name,
                    schedule.name,
                )
                return
            else:
                raise e



"""Unique run index [6e4eb89f632d].

Revision ID: 6e4eb89f632d
Revises: 5c0a1c787128
Create Date: 2025-12-03 17:27:32.828004

"""

from typing import Any

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6e4eb89f632d"
down_revision = "5c0a1c787128"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    with op.batch_alter_table("pipeline", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("run_count", sa.Integer(), nullable=True)
        )

    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.add_column(sa.Column("index", sa.Integer(), nullable=True))

    connection = op.get_bind()
    meta = sa.MetaData()
    meta.reflect(bind=connection, only=("pipeline_run", "pipeline"))
    run_table = sa.Table("pipeline_run", meta)
    pipeline_table = sa.Table("pipeline", meta)

    # Some very old runs (version <0.34.0) might not have a pipeline ID in the
    # rare case where the associated pipeline was deleted after the migration
    # with revision `8ad841ad9bfe`.
    connection.execute(
        sa.update(run_table)
        .where(run_table.c.pipeline_id.is_(None))
        .values(index=0)
    )

    result = connection.execute(
        sa.select(
            run_table.c.id,
            run_table.c.pipeline_id,
        )
        .where(run_table.c.pipeline_id.is_not(None))
        .order_by(run_table.c.pipeline_id, run_table.c.created, run_table.c.id)
    ).fetchall()

    current_pipeline_id = None
    index_within_pipeline = 0
    run_updates: list[dict[str, Any]] = []
    run_counts: dict[str, int] = {}
    for row in result:
        pipeline_id = row.pipeline_id
        if pipeline_id != current_pipeline_id:
            current_pipeline_id = pipeline_id
            index_within_pipeline = 1
        else:
            index_within_pipeline += 1
        run_updates.append({"id_": row.id, "index": index_within_pipeline})
        run_counts[pipeline_id] = index_within_pipeline

    update_batch_size = 10000
    if run_updates:
        update_statement = (
            sa.update(run_table)
            .where(run_table.c.id == sa.bindparam("id_"))
            .values(index=sa.bindparam("index"))
        )

        for start in range(0, len(run_updates), update_batch_size):
            batch = run_updates[start : start + update_batch_size]
            if batch:
                connection.execute(update_statement, batch)

    if run_counts:
        pipeline_updates = [
            {"id_": pipeline_id, "run_count": run_count}
            for pipeline_id, run_count in run_counts.items()
        ]
        update_statement = (
            sa.update(pipeline_table)
            .where(pipeline_table.c.id == sa.bindparam("id_"))
            .values(run_count=sa.bindparam("run_count"))
        )
        for start in range(0, len(pipeline_updates), update_batch_size):
            batch = pipeline_updates[start : start + update_batch_size]
            if batch:
                connection.execute(update_statement, batch)

    # Set run_count to 0 for pipelines without any runs
    connection.execute(
        sa.update(pipeline_table)
        .where(pipeline_table.c.run_count.is_(None))
        .values(run_count=0)
    )

    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.alter_column(
            "index", existing_type=sa.Integer(), nullable=False
        )

    with op.batch_alter_table("pipeline", schema=None) as batch_op:
        batch_op.alter_column(
            "run_count", existing_type=sa.Integer(), nullable=False
        )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    with op.batch_alter_table("pipeline_run", schema=None) as batch_op:
        batch_op.drop_column("number")

    with op.batch_alter_table("pipeline", schema=None) as batch_op:
        batch_op.drop_column("run_count")
